# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020, 2021 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from argparse import SUPPRESS, Action, ArgumentParser
from dataclasses import dataclass, field
from pathlib import Path
from sys import stdout
from typing import TYPE_CHECKING

from .interface_generator import (
    generate_py_file,
    interfaces_from_file,
    interfaces_from_str,
)

if TYPE_CHECKING:
    from typing import Optional

    from .interface_generator import DbusInterfaceIntrospection


@dataclass
class RenameMember:
    new_name: Optional[str] = None
    current_arg: Optional[str] = None
    arg_renames: dict[str, str] = field(default_factory=dict)


@dataclass
class RenameInterface:
    new_name: Optional[str] = None
    current_member: Optional[RenameMember] = None
    methods: dict[str, RenameMember] = field(default_factory=dict)
    properties: dict[str, RenameMember] = field(default_factory=dict)
    signals: dict[str, RenameMember] = field(default_factory=dict)


@dataclass
class RenameRoot:
    current_interface: Optional[RenameInterface] = None
    interfaces: dict[str, RenameInterface] = field(default_factory=dict)


rename_root = RenameRoot()


# def rename_args(member_rename):
#     ...


def rename_members(
    interface: DbusInterfaceIntrospection,
    interface_rename: RenameInterface,
) -> None:
    for m_member in interface.methods:
        m_rename = interface_rename.methods.get(m_member.method_name)
        if m_rename is None:
            continue

        if m_rename.new_name is not None:
            m_member.python_name = m_rename.new_name

    for p_member in interface.properties:
        p_rename = interface_rename.properties.get(p_member.method_name)
        if p_rename is None:
            continue

        if p_rename.new_name is not None:
            p_member.python_name = p_rename.new_name

    for s_member in interface.signals:
        s_rename = interface_rename.signals.get(s_member.method_name)
        if s_rename is None:
            continue

        if s_rename.new_name is not None:
            s_member.python_name = s_rename.new_name


def rename_interfaces(
    interfaces: list[DbusInterfaceIntrospection]
) -> None:
    for interface in interfaces:
        dbus_interface_name = interface.interface_name
        this_interface_rename = rename_root.interfaces.get(dbus_interface_name)
        if this_interface_rename is None:
            continue

        if this_interface_rename.new_name is not None:
            interface.python_name = this_interface_rename.new_name

        rename_members(interface, this_interface_rename)


def run_gen_from_connection(
    connection_name: str,
    object_paths: list[str],
    system: bool,
    imports_header: bool,
    do_async: bool,
) -> None:
    connection_name = connection_name
    object_paths = object_paths

    from .dbus_proxy_sync_interfaces import DbusInterfaceCommon

    if system:
        from .default_bus import set_default_bus
        from .sd_bus_internals import sd_bus_open_system
        set_default_bus(sd_bus_open_system())

    interfaces: list[DbusInterfaceIntrospection] = []
    for object_path in object_paths:
        connection = DbusInterfaceCommon(connection_name, object_path)
        itrospection = connection.dbus_introspect()
        interfaces.extend(interfaces_from_str(itrospection))

    rename_interfaces(interfaces)

    stdout.write(
        generate_py_file(
            interfaces,
            imports_header,
            do_async,
        )
    )


def run_gen_from_file(
    filenames: list[str],
    imports_header: bool,
    do_async: bool,
) -> None:
    interfaces: list[DbusInterfaceIntrospection] = []

    for file in filenames:
        interfaces.extend(interfaces_from_file(file))

    rename_interfaces(interfaces)

    stdout.write(
        generate_py_file(
            interfaces,
            imports_header,
            do_async,
        )
    )


class ActionSelectInterface(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: object,
        values: object,
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError(
                f"Expected --select-interface to be string, got {values!r}"
            )

        interface_rename = rename_root.interfaces.get(values)

        if interface_rename is None:
            interface_rename = RenameInterface()
            rename_root.interfaces[values] = interface_rename

        rename_root.current_interface = interface_rename


class ActionSelectMethod(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: object,
        values: object,
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError(
                f"Expected --select-method to be string, got {values!r}"
            )

        current_interface = rename_root.current_interface
        if current_interface is None:
            raise ValueError(
                "No D-Bus interface selected. "
                "Use --select-interface option."
            )

        method_rename = current_interface.methods.get(values)

        if method_rename is None:
            method_rename = RenameMember()
            current_interface.methods[values] = method_rename

        current_interface.current_member = method_rename


class ActionSelectProperty(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: object,
        values: object,
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError(
                f"Expected --select-property to be string, got {values!r}"
            )

        current_interface = rename_root.current_interface
        if current_interface is None:
            raise ValueError(
                "No D-Bus interface selected. "
                "Use --select-interface option."
            )

        property_rename = current_interface.properties.get(values)

        if property_rename is None:
            property_rename = RenameMember()
            current_interface.properties[values] = property_rename

        current_interface.current_member = property_rename


class ActionSelectSignal(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: object,
        values: object,
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError(
                f"Expected --select-signal to be string, got {values!r}"
            )

        current_interface = rename_root.current_interface
        if current_interface is None:
            raise ValueError(
                "No D-Bus interface selected. "
                "Use --select-interface option."
            )

        signal_rename = current_interface.signals.get(values)

        if signal_rename is None:
            signal_rename = RenameMember()
            current_interface.signals[values] = signal_rename

        current_interface.current_member = signal_rename


class ActionSetName(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: object,
        values: object,
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError(
                f"Expected --set-name to be string, got {values!r}"
            )

        current_interface = rename_root.current_interface
        current_member = (
            current_interface.current_member
            if current_interface is not None
            else None
        )

        if current_member is not None:
            current_member.new_name = values
            return

        if current_interface is not None:
            current_interface.new_name = values
            return

        raise ValueError(
            "No D-Bus element to rename. "
            "Use --select-* options to select element."
        )


def generator_main(args: Optional[list[str]] = None) -> None:

    main_arg_parser = ArgumentParser(
        prog="sdbus",
    )
    subparsers = main_arg_parser.add_subparsers(
        required=True,
        title="subcommands",
    )

    generate_from_file_parser = subparsers.add_parser('gen-from-file')
    generate_from_file_parser.set_defaults(func=run_gen_from_file)

    generate_from_connection = subparsers.add_parser('gen-from-connection')
    generate_from_connection.set_defaults(func=run_gen_from_connection)

    # Common options
    for subparser in (generate_from_file_parser, generate_from_connection):
        subparser.add_argument(
            "--no-imports-header", action="store_false",
            dest="imports_header",
            help="Do NOT include 'import' header",
        )
        subparser.add_argument(
            "--imports-header", action="store_true", default=True,
            dest="imports_header",
            help="Include 'import' header (default)",
        )

        subparser.add_argument(
            "--async", action="store_true", default=True,
            dest="do_async",
            help="Generate async interfaces (default)",
        )
        subparser.add_argument(
            "--block", action="store_false",
            dest="do_async",
            help="Generate blocking interfaces",
        )
        subparser.add_argument(
            "--select-interface",
            action=ActionSelectInterface,
            default=SUPPRESS,
            help="Select D-Bus interface to adjust"
        )
        subparser.add_argument(
            "--select-method",
            action=ActionSelectMethod,
            default=SUPPRESS,
            help="Select D-Bus method to adjust"
        )
        subparser.add_argument(
            "--select-property",
            action=ActionSelectProperty,
            default=SUPPRESS,
            help="Select D-Bus property to adjust"
        )
        subparser.add_argument(
            "--select-signal",
            action=ActionSelectSignal,
            default=SUPPRESS,
            help="Select D-Bus signal to adjust"
        )
        subparser.add_argument(
            "--set-name",
            action=ActionSetName,
            default=SUPPRESS,
            help="Select D-Bus interface to adjust"
        )

    generate_from_file_parser.add_argument(
        'filenames', type=Path, nargs='+',
        help="Paths to interface XML introspection files"
    )

    generate_from_connection.add_argument(
        'connection_name',
        help=(
            'Name of the service connection to extract introspection. '
            "For example, 'org.freedesktop.systemd1' for systemd."
        )
    )
    generate_from_connection.add_argument(
        'object_paths',
        nargs='+',
        help=(
            'Object paths that the introspection will be extracted. '
            'One or more.'
        )
    )
    generate_from_connection.add_argument(
        '--system',
        help='Use system D-Bus instead of session.',
        action='store_true',
    )

    args_dict = vars(main_arg_parser.parse_args(args))
    func = args_dict.pop("func")
    func(**args_dict)


if __name__ == "__main__":
    generator_main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from typing import TYPE_CHECKING

from .dbus_proxy_sync_interface_base import DbusInterfaceBase
from .dbus_proxy_sync_method import dbus_method

if TYPE_CHECKING:
    from typing import Any, Literal


class DbusPeerInterface(
    DbusInterfaceBase,
    interface_name='org.freedesktop.DBus.Peer',
    serving_enabled=False,
):

    @dbus_method(method_name='Ping')
    def dbus_ping(self) -> None:
        raise NotImplementedError

    @dbus_method(method_name='GetMachineId')
    def dbus_machine_id(self) -> str:
        raise NotImplementedError


class DbusIntrospectable(
    DbusInterfaceBase,
    interface_name='org.freedesktop.DBus.Introspectable',
    serving_enabled=False,
):

    @dbus_method(method_name='Introspect')
    def dbus_introspect(self) -> str:
        raise NotImplementedError


class DbusPropertiesInterface(
    DbusInterfaceBase,
    interface_name='org.freedesktop.DBus.Properties',
    serving_enabled=False,
):
    @dbus_method('s', 'a{sv}', method_name='GetAll')
    def _properties_get_all(
            self, interface_name: str) -> dict[str, tuple[str, Any]]:
        raise NotImplementedError

    def properties_get_all_dict(
            self,
            on_unknown_member: Literal['error', 'ignore', 'reuse'] = 'error',
    ) -> dict[str, Any]:
        properties: dict[str, Any] = {}

        for interface_name, meta in self._dbus_iter_interfaces_meta():
            if not meta.serving_enabled:
                continue

            dbus_properties_data = self._properties_get_all(interface_name)
            for member_name, variant in dbus_properties_data.items():
                try:
                    python_name = meta.dbus_member_to_python_attr[member_name]
                except KeyError:
                    if on_unknown_member == 'error':
                        raise
                    elif on_unknown_member == 'ignore':
                        continue
                    elif on_unknown_member == 'reuse':
                        python_name = member_name
                    else:
                        raise ValueError

                properties[python_name] = variant[1]

        return properties


class DbusInterfaceCommon(
        DbusPropertiesInterface,
        DbusIntrospectable,
        DbusPeerInterface):
    ...


class DbusObjectManagerInterface(
    DbusInterfaceCommon,
    interface_name='org.freedesktop.DBus.ObjectManager',
    serving_enabled=False,
):
    @dbus_method(result_signature='a{oa{sa{sv}}}')
    def get_managed_objects(
            self) -> dict[str, dict[str, dict[str, Any]]]:
        raise NotImplementedError



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import get_running_loop, wait_for
from typing import Any

from sdbus.dbus_common_elements import DbusLocalObjectMeta
from sdbus.exceptions import DbusFailedError
from sdbus.unittest import IsolatedDbusTestCase

from sdbus import (
    DbusInterfaceCommonAsync,
    dbus_method_async,
    dbus_property_async,
    request_default_bus_name_async,
)

HELLO_WORLD = 'Hello, world!'


class DbusDerivePropertydError(DbusFailedError):
    dbus_error_name = 'org.example.PropertyError'


class IndependentError(Exception):
    ...


GOOD_STR = 'Good'


class InterfaceWithErrors(
    DbusInterfaceCommonAsync,
    interface_name='org.example.errors',
):
    @dbus_property_async('s')
    def indep_err_getter(self) -> str:
        raise IndependentError

    @dbus_property_async('s')
    def derrive_err_getter(self) -> str:
        raise DbusDerivePropertydError

    @dbus_method_async(result_signature='s')
    async def hello_error(self) -> str:
        raise AttributeError

    @dbus_method_async(result_signature='s')
    async def hello_world(self) -> str:
        return HELLO_WORLD

    @dbus_property_async('s')
    def indep_err_setable(self) -> str:
        return GOOD_STR

    @indep_err_setable.setter
    def indep_err_setter(self, new_value: str) -> None:
        raise IndependentError

    @dbus_property_async('s')
    def derrive_err_settable(self) -> str:
        return GOOD_STR

    @derrive_err_settable.setter
    def derrive_err_setter(self, new_value: str) -> None:
        raise DbusDerivePropertydError


class TestLowLevelErrors(IsolatedDbusTestCase):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()

        await request_default_bus_name_async('org.test')
        self.test_object = InterfaceWithErrors()
        self.test_object.export_to_dbus('/')

        self.test_object_connection = InterfaceWithErrors.new_proxy(
            'org.test', '/')

        loop = get_running_loop()

        def silence_exceptions(*args: Any, **kwrags: Any) -> None:
            ...

        loop.set_exception_handler(silence_exceptions)

    async def test_property_getter_independent_error(self) -> None:
        with self.assertRaises(DbusFailedError) as cm:
            await wait_for(
                self.test_object_connection.indep_err_getter.get_async(),
                timeout=1,
            )

        should_be_dbus_failed = cm.exception
        self.assertIs(should_be_dbus_failed.__class__, DbusFailedError)

        await self.test_object_connection.hello_world()

    async def test_property_getter_derived_error(self) -> None:
        with self.assertRaises(DbusDerivePropertydError):
            await wait_for(
                self.test_object_connection.derrive_err_getter.get_async(),
                timeout=1,
            )

        await self.test_object_connection.hello_world()

    async def test_property_setter_independent_error(self) -> None:

        self.assertEqual(
            await wait_for(
                self.test_object_connection.indep_err_setable.get_async(),
                timeout=1,
            ),
            GOOD_STR,
        )

        with self.assertRaises(DbusFailedError) as cm:
            await wait_for(
                self.test_object_connection.
                indep_err_setable.set_async('Test'),
                timeout=1,
            )

        should_be_dbus_failed = cm.exception
        self.assertIs(should_be_dbus_failed.__class__, DbusFailedError)

        await self.test_object_connection.hello_world()

    async def test_property_setter_derived_error(self) -> None:

        self.assertEqual(
            await wait_for(
                self.test_object_connection.derrive_err_settable.get_async(),
                timeout=1,
            ),
            GOOD_STR,
        )

        with self.assertRaises(DbusDerivePropertydError):
            await wait_for(
                self.test_object_connection.
                derrive_err_settable.set_async('Test'),
                timeout=1,
            )

        await self.test_object_connection.hello_world()

    async def test_property_callback_error(self) -> None:
        dbus_local_meta = self.test_object._dbus
        if not isinstance(dbus_local_meta, DbusLocalObjectMeta):
            raise TypeError
        interface = dbus_local_meta.activated_interfaces[0]
        interface.property_get_dict.pop(b'DerriveErrSettable')

        with self.assertRaises(DbusFailedError):
            await wait_for(
                self.test_object_connection.derrive_err_settable,
                timeout=1,
            )

    async def test_method_callback_error(self) -> None:
        TEST_KEY = b'HelloWorld'
        dbus_local_meta = self.test_object._dbus
        if not isinstance(dbus_local_meta, DbusLocalObjectMeta):
            raise TypeError
        interface = dbus_local_meta.activated_interfaces[0]
        interface.method_dict.pop(TEST_KEY)

        with self.assertRaises(DbusFailedError):
            await wait_for(
                self.test_object_connection.hello_world(),
                timeout=1,
            )

        interface.method_dict[TEST_KEY] = None

        with self.assertRaises(TypeError):
            await wait_for(
                self.test_object_connection.hello_world(),
                timeout=1,
            )

        def test_raise(*args: Any, **kwargs: Any) -> None:
            raise DbusDerivePropertydError

        interface.method_dict[TEST_KEY] = test_raise
        with self.assertRaises(DbusDerivePropertydError):
            await wait_for(
                self.test_object_connection.hello_world(),
                timeout=1,
            )



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from collections.abc import Callable
from copy import copy
from itertools import chain
from typing import TYPE_CHECKING, Any, cast
from warnings import warn
from weakref import WeakKeyDictionary, WeakValueDictionary

from .dbus_common_elements import (
    DbusClassMeta,
    DbusInterfaceMetaCommon,
    DbusLocalObjectMeta,
    DbusMemberAsync,
    DbusMemberCommon,
    DbusMemberSync,
    DbusMethodOverride,
    DbusPropertyOverride,
    DbusRemoteObjectMeta,
)
from .dbus_proxy_async_method import DbusLocalMethodAsync, DbusMethodAsync
from .dbus_proxy_async_property import (
    DbusLocalPropertyAsync,
    DbusPropertyAsync,
)
from .dbus_proxy_async_signal import DbusLocalSignalAsync, DbusSignalAsync
from .default_bus import get_default_bus
from .sd_bus_internals import SdBusInterface

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Optional, TypeVar, Union

    from .sd_bus_internals import SdBus, SdBusSlot

    T = TypeVar('T')
    Self = TypeVar('Self', bound="DbusInterfaceBaseAsync")
    DbusOverride = Union[DbusMethodOverride[T], DbusPropertyOverride[T]]


DBUS_CLASS_TO_META: WeakKeyDictionary[
    type, DbusClassMeta] = WeakKeyDictionary()
DBUS_INTERFACE_NAME_TO_CLASS: WeakValueDictionary[
    str, DbusInterfaceMetaAsync] = WeakValueDictionary()


class DbusInterfaceMetaAsync(DbusInterfaceMetaCommon):

    @staticmethod
    def _process_dbus_method_override(
        override_attr_name: str,
        override: DbusMethodOverride[T],
        mro_dbus_elements: dict[str, DbusMemberAsync],
    ) -> DbusMethodAsync:
        try:
            original_dbus_method = mro_dbus_elements[override_attr_name]
        except KeyError:
            raise ValueError(
                f"No D-Bus method {override_attr_name!r} found "
                f"to override."
            )

        if not isinstance(original_dbus_method, DbusMethodAsync):
            raise TypeError(
                f"Expected {DbusMethodAsync!r} got {original_dbus_method!r} "
                f"under name {override_attr_name!r}"
            )

        new_method = copy(original_dbus_method)
        new_method.original_method = (
            override.override_method  # type: ignore[assignment]
        )
        return new_method

    @staticmethod
    def _process_dbus_property_override(
        override_attr_name: str,
        override: DbusPropertyOverride[T],
        mro_dbus_elements: dict[str, DbusMemberAsync],
    ) -> DbusPropertyAsync[Any]:
        try:
            original_property = mro_dbus_elements[override_attr_name]
        except KeyError:
            raise ValueError(
                f"No D-Bus property {override_attr_name!r} found "
                f"to override."
            )

        if not isinstance(original_property, DbusPropertyAsync):
            raise TypeError(
                f"Expected {DbusMethodAsync!r} got {original_property!r} "
                f"under name {override_attr_name!r}"
            )

        new_property = copy(original_property)
        new_property.property_getter = cast(
            Callable[[DbusInterfaceBaseAsync], Any],
            override.getter_override
        )
        if override.setter_override is not None:
            new_property.property_setter = override.setter_override
            new_property.property_setter_is_public = override.is_setter_public

        return new_property

    @classmethod
    def _check_collisions(
        cls,
        new_class_name: str,
        namespace: dict[str, Any],
        mro_dbus_elements: dict[str, DbusMemberAsync],
    ) -> None:

        possible_collisions = namespace.keys() & mro_dbus_elements.keys()
        new_overrides: dict[str, DbusMemberAsync] = {}

        for attr_name, attr in namespace.items():
            if isinstance(attr, DbusMethodOverride):
                new_overrides[attr_name] = cls._process_dbus_method_override(
                    attr_name,
                    attr,
                    mro_dbus_elements,
                )
                possible_collisions.remove(attr_name)
            elif isinstance(attr, DbusPropertyOverride):
                new_overrides[attr_name] = cls._process_dbus_property_override(
                    attr_name,
                    attr,
                    mro_dbus_elements,
                )
                possible_collisions.remove(attr_name)
            else:
                continue

        if possible_collisions:
            raise ValueError(
                f"Interface {new_class_name!r} redefines reserved "
                f"D-Bus attribute names: {possible_collisions!r}"
            )

        namespace.update(new_overrides)

    @staticmethod
    def _extract_dbus_elements(
        dbus_class: type,
        dbus_meta: DbusClassMeta,
    ) -> dict[str, DbusMemberAsync]:
        dbus_elements_map: dict[str, DbusMemberAsync] = {}

        for attr_name in dbus_meta.python_attr_to_dbus_member.keys():
            dbus_element = dbus_class.__dict__.get(attr_name)
            if not isinstance(dbus_element, DbusMemberAsync):
                raise TypeError(
                    f"Expected async D-Bus element, got {dbus_element!r} "
                    f"in class {dbus_class!r}"
                )

            dbus_elements_map[attr_name] = dbus_element

        return dbus_elements_map

    @classmethod
    def _map_mro_dbus_elements(
        cls,
        new_class_name: str,
        base_classes: Iterable[type],
    ) -> dict[str, DbusMemberAsync]:
        all_python_dbus_map: dict[str, DbusMemberAsync] = {}
        possible_collisions: set[str] = set()

        for c in base_classes:
            dbus_meta = DBUS_CLASS_TO_META.get(c)
            if dbus_meta is None:
                continue

            base_dbus_elements = cls._extract_dbus_elements(c, dbus_meta)

            possible_collisions.update(
                base_dbus_elements.keys() & all_python_dbus_map.keys()
            )

            all_python_dbus_map.update(
                base_dbus_elements
            )

        if possible_collisions:
            raise ValueError(
                f"Interface {new_class_name!r} has a reserved D-Bus "
                f"attribute name collision: {possible_collisions!r}"
            )

        return all_python_dbus_map

    @staticmethod
    def _map_dbus_elements(
        attr_name: str,
        attr: Any,
        meta: DbusClassMeta,
        interface_name: str,
    ) -> None:
        if not isinstance(attr, DbusMemberCommon):
            return

        if isinstance(attr, DbusMemberSync):
            raise TypeError(
                "Can't mix blocking methods in "
                f"async interface: {attr_name!r}"
            )

        if attr.interface_name != interface_name:
            return

        if isinstance(attr, DbusMethodAsync):
            meta.dbus_member_to_python_attr[attr.method_name] = attr_name
            meta.python_attr_to_dbus_member[attr_name] = attr.method_name
        elif isinstance(attr, DbusPropertyAsync):
            meta.dbus_member_to_python_attr[attr.property_name] = attr_name
            meta.python_attr_to_dbus_member[attr_name] = attr.property_name
        elif isinstance(attr, DbusSignalAsync):
            meta.dbus_member_to_python_attr[attr.signal_name] = attr_name
            meta.python_attr_to_dbus_member[attr_name] = attr.signal_name
        else:
            raise TypeError(f"Unknown D-Bus element: {attr!r}")

    def __new__(cls, name: str,
                bases: tuple[type, ...],
                namespace: dict[str, Any],
                interface_name: Optional[str] = None,
                serving_enabled: bool = True,
                ) -> DbusInterfaceMetaAsync:

        if interface_name in DBUS_INTERFACE_NAME_TO_CLASS:
            raise ValueError(
                f"D-Bus interface of the name {interface_name!r} was "
                "already created."
            )

        all_mro_bases: set[type[Any]] = set(
            chain.from_iterable(c.__mro__ for c in bases)
        )
        reserved_dbus_map = cls._map_mro_dbus_elements(
            name, all_mro_bases,
        )
        cls._check_collisions(name, namespace, reserved_dbus_map)

        new_cls = super().__new__(
            cls, name, bases, namespace,
            interface_name,
            serving_enabled,
        )

        if interface_name is not None:
            dbus_class_meta = DbusClassMeta(interface_name, serving_enabled)
            DBUS_CLASS_TO_META[new_cls] = dbus_class_meta
            DBUS_INTERFACE_NAME_TO_CLASS[interface_name] = new_cls

            for attr_name, attr in namespace.items():
                cls._map_dbus_elements(
                    attr_name,
                    attr,
                    dbus_class_meta,
                    interface_name,
                )

        return new_cls


class DbusInterfaceBaseAsync(metaclass=DbusInterfaceMetaAsync):

    def __init__(self) -> None:
        self._dbus: Union[
            DbusRemoteObjectMeta, DbusLocalObjectMeta] = DbusLocalObjectMeta()

    @classmethod
    def _dbus_iter_interfaces_meta(
        cls,
    ) -> Iterator[tuple[str, DbusClassMeta]]:

        for base in reversed(cls.__mro__):
            meta = DBUS_CLASS_TO_META.get(base)
            if meta is None:
                continue

            yield meta.interface_name, meta

    async def start_serving(self,
                            object_path: str,
                            bus: Optional[SdBus] = None,
                            ) -> None:

        warn("start_serving is deprecated in favor of export_to_dbus",
             DeprecationWarning)
        self.export_to_dbus(object_path, bus)

    def _dbus_on_no_members_exported(self) -> None:
        raise ValueError("No D-Bus interfaces were exported")

    def export_to_dbus(
        self,
        object_path: str,
        bus: Optional[SdBus] = None,
    ) -> DbusExportHandle:
        local_object_meta = self._dbus
        if isinstance(local_object_meta, DbusRemoteObjectMeta):
            raise RuntimeError("Cannot export D-Bus proxies.")

        # TODO: Being able to serve multiple buses and object
        if local_object_meta.attached_bus is not None:
            raise RuntimeError(
                "Object already exported. "
                "This limitation should be fixed in future version."
            )

        if bus is None:
            bus = get_default_bus()

        local_object_meta.attached_bus = bus
        local_object_meta.serving_object_path = object_path

        for interface_name, meta in self._dbus_iter_interfaces_meta():
            if not meta.serving_enabled:
                continue

            new_interface = SdBusInterface()

            for python_attr, dbus_member in (
                meta.python_attr_to_dbus_member.items()
            ):
                dbus_something = getattr(self, python_attr)
                if isinstance(dbus_something, DbusLocalMethodAsync):
                    new_interface.add_method(
                        dbus_something.dbus_method.method_name,
                        dbus_something.dbus_method.input_signature,
                        dbus_something.dbus_method.input_args_names,
                        dbus_something.dbus_method.result_signature,
                        dbus_something.dbus_method.result_args_names,
                        dbus_something.dbus_method.flags,
                        dbus_something._dbus_reply_call,
                    )
                elif isinstance(dbus_something, DbusLocalPropertyAsync):
                    getter = dbus_something._dbus_reply_get
                    dbus_property = dbus_something.dbus_property

                    if (
                        dbus_property.property_setter is not None
                        and
                        dbus_property.property_setter_is_public
                    ):
                        setter = dbus_something._dbus_reply_set
                    else:
                        setter = None

                    new_interface.add_property(
                        dbus_property.property_name,
                        dbus_property.property_signature,
                        getter,
                        setter,
                        dbus_property.flags,
                    )
                elif isinstance(dbus_something, DbusLocalSignalAsync):
                    new_interface.add_signal(
                        dbus_something.dbus_signal.signal_name,
                        dbus_something.dbus_signal.signal_signature,
                        dbus_something.dbus_signal.args_names,
                        dbus_something.dbus_signal.flags,
                    )
                else:
                    raise TypeError(
                        f"Expected D-Bus element, got: {dbus_something!r}"
                    )

            bus.add_interface(new_interface, object_path, interface_name)
            local_object_meta.activated_interfaces.append(new_interface)

        if not local_object_meta.activated_interfaces:
            self._dbus_on_no_members_exported()

        return DbusExportHandle(local_object_meta)

    def _connect(
        self,
        service_name: str,
        object_path: str,
        bus: Optional[SdBus] = None,
    ) -> None:
        self._proxify(
            service_name,
            object_path,
            bus,
        )

    def _proxify(
        self,
        service_name: str,
        object_path: str,
        bus: Optional[SdBus] = None,
    ) -> None:

        self._dbus = DbusRemoteObjectMeta(
            service_name,
            object_path,
            bus,
        )

    @classmethod
    def new_connect(
        cls: type[Self],
        service_name: str,
        object_path: str,
        bus: Optional[SdBus] = None,
    ) -> Self:
        warn(
            ("new_connect is deprecated in favor of equivalent new_proxy."
             "Will be removed in version 1.0.0"),
            DeprecationWarning,
        )
        new_object = cls.__new__(cls)
        new_object._proxify(service_name, object_path, bus)
        return new_object

    @classmethod
    def new_proxy(
        cls: type[Self],
        service_name: str,
        object_path: str,
        bus: Optional[SdBus] = None,
    ) -> Self:

        new_object = cls.__new__(cls)
        new_object._proxify(service_name, object_path, bus)
        return new_object


class DbusExportHandle:
    def __init__(self, local_meta: DbusLocalObjectMeta):
        self._tasks = local_meta.tasks
        self._dbus_slots: list[SdBusSlot] = [
            i.slot
            for i in local_meta.activated_interfaces
            if i.slot is not None
        ]

    async def __aenter__(self) -> DbusExportHandle:
        return self

    def __enter__(self) -> DbusExportHandle:
        return self

    def __exit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any,
    ) -> None:
        self.stop()

    async def __aexit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any,
    ) -> None:
        self.stop()

    def stop(self) -> None:
        for task in self._tasks:
            task.cancel("D-Bus export stopped")

        for slot in self._dbus_slots:
            slot.close()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import get_running_loop
from typing import TYPE_CHECKING

from .sd_bus_internals import (
    DbusPropertyConstFlag,
    DbusPropertyEmitsChangeFlag,
    DbusPropertyEmitsInvalidationFlag,
    DbusPropertyExplicitFlag,
)

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from typing import Any, Literal


PROPERTY_FLAGS_MASK = (
    DbusPropertyConstFlag | DbusPropertyEmitsChangeFlag |
    DbusPropertyEmitsInvalidationFlag | DbusPropertyExplicitFlag
)


def count_bits(i: int) -> int:
    return bin(i).count('1')


def _is_property_flags_correct(flags: int) -> bool:
    num_of_flag_bits = count_bits(PROPERTY_FLAGS_MASK & flags)
    return (0 <= num_of_flag_bits <= 1)


def _snake_case_to_camel_case_gen(snake: str) -> Iterator[str]:
    char_iter = iter(snake)
    # Name starting with upper case letter
    try:
        first_char = next(char_iter)
    except StopIteration:
        return

    yield first_char.upper()

    upper_next_one = False
    for c in char_iter:
        # Every under score remove and uppercase next one
        if c != '_':
            if upper_next_one:
                yield c.upper()
                upper_next_one = False
            else:
                yield c
        else:
            upper_next_one = True


def snake_case_to_camel_case(snake: str) -> str:
    return "".join(_snake_case_to_camel_case_gen(snake))


def _check_sync_in_async_env() -> bool:
    try:
        get_running_loop()
        return False
    except RuntimeError:
        return True


def _parse_properties_vardict(
        properties_name_map: Mapping[str, str],
        properties_vardict: dict[str, tuple[str, Any]],
        on_unknown_member: Literal['error', 'ignore', 'reuse'],
) -> dict[str, Any]:

    properties_translated: dict[str, Any] = {}

    for member_name, variant in properties_vardict.items():
        try:
            python_name = properties_name_map[member_name]
        except KeyError:
            if on_unknown_member == 'error':
                raise
            elif on_unknown_member == 'ignore':
                continue
            elif on_unknown_member == 'reuse':
                python_name = member_name
            else:
                raise ValueError

        properties_translated[python_name] = variant[1]

    return properties_translated



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import Future
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Optional, Union

    DbusBasicTypes = Union[str, int, bytes, float, Any]
    DbusStructType = tuple[DbusBasicTypes, ...]
    DbusDictType = dict[DbusBasicTypes, DbusBasicTypes]
    DbusVariantType = tuple[str, DbusStructType]
    DbusListType = list[DbusBasicTypes]
    DbusCompleteTypes = Union[DbusBasicTypes, DbusStructType,
                              DbusDictType, DbusVariantType, DbusListType]

__STUB_ERROR = (
    'Typing stub. You should never see this '
    'error unless the actual module failed to load. '
    'Check your installation.'
)


class SdBusSlot:
    """Holds reference to SdBus slot"""

    def close(self) -> None:
        raise NotImplementedError(__STUB_ERROR)


class SdBusInterface:
    slot: Optional[SdBusSlot]
    method_list: list[object]
    method_dict: dict[bytes, object]
    property_list: list[object]
    property_get_dict: dict[bytes, object]
    property_set_dict: dict[bytes, object]
    signal_list: list[object]

    def add_method(
        self,
        member_name: str,
        signature: str, input_args_names: Sequence[str],
        result_signature: str, result_args_names: Sequence[str],
        flags: int,
        callback: Callable[[SdBusMessage], None], /
    ) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def add_property(
        self,
        property_name: str,
        property_signature: str,
        get_function: Callable[[SdBusMessage], Any],
        set_function: Optional[Callable[[SdBusMessage], None]],
        flags: int, /
    ) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def add_signal(
        self,
        signal_name: str,
        signal_signature: str,
        signal_args_names: Sequence[str],
        flags: int, /
    ) -> None:
        raise NotImplementedError(__STUB_ERROR)


class SdBusMessage:
    def append_data(self, signature: str, *args: DbusCompleteTypes) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def open_container(self, container_type: str,
                       container_signature: str, /) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def close_container(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def enter_container(self, container_type: str,
                        container_signature: str, /) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def exit_container(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def dump(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def seal(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def get_contents(self
                     ) -> tuple[DbusCompleteTypes, ...]:
        raise NotImplementedError(__STUB_ERROR)

    def create_reply(self) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def create_error_reply(
            self,
            error_name: str,
            error_message: str, /) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def send(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def parse_to_tuple(self) -> tuple[Any, ...]:
        raise NotImplementedError(__STUB_ERROR)

    expect_reply: bool = False
    destination: Optional[str] = None
    path: Optional[str] = None
    interface: Optional[str] = None
    member: Optional[str] = None
    sender: Optional[str] = None


class SdBus:
    def call(self, message: SdBusMessage, /) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def call_async(
            self, message: SdBusMessage,
            /) -> Future[SdBusMessage]:
        raise NotImplementedError(__STUB_ERROR)

    def process(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def get_fd(self) -> int:
        raise NotImplementedError(__STUB_ERROR)

    def new_method_call_message(
            self,
            destination_name: str, object_path: str,
            interface_name: str, member_name: str,
            /) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def new_property_get_message(
            self,
            destination_service_name: str, object_path: str,
            interface_name: str, member_name: str,
            /) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def new_property_set_message(
            self,
            destination_service_name: str, object_path: str,
            interface_name: str, member_name: str,
            /) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def new_signal_message(
            self,
            object_path: str,
            interface_name: str,
            member_name: str,
            /) -> SdBusMessage:
        raise NotImplementedError(__STUB_ERROR)

    def add_interface(self, new_interface: SdBusInterface,
                      object_path: str, interface_name: str, /) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def match_signal_async(
        self,
        senders_name: Optional[str], object_path: Optional[str],
        interface_name: Optional[str], member_name: Optional[str],
        callback: Callable[[SdBusMessage], None], /
    ) -> Future[SdBusSlot]:
        raise NotImplementedError(__STUB_ERROR)

    def request_name_async(self, name: str, flags: int, /) -> Future[None]:
        raise NotImplementedError(__STUB_ERROR)

    def request_name(self, name: str, flags: int, /) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def add_object_manager(self, path: str, /) -> SdBusSlot:
        raise NotImplementedError(__STUB_ERROR)

    def emit_object_added(self, path: str, /) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def emit_object_removed(self, path: str, /) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def close(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    def start(self) -> None:
        raise NotImplementedError(__STUB_ERROR)

    address: Optional[str] = None
    method_call_timeout_usec: int = 0


def sd_bus_open() -> SdBus:
    raise NotImplementedError(__STUB_ERROR)


def sd_bus_open_user() -> SdBus:
    raise NotImplementedError(__STUB_ERROR)


def sd_bus_open_system() -> SdBus:
    raise NotImplementedError(__STUB_ERROR)


def sd_bus_open_system_remote(host: str, /) -> SdBus:
    raise NotImplementedError(__STUB_ERROR)


def sd_bus_open_user_machine(machine: str, /) -> SdBus:
    raise NotImplementedError(__STUB_ERROR)


def sd_bus_open_system_machine(machine: str, /) -> SdBus:
    raise NotImplementedError(__STUB_ERROR)


def encode_object_path(prefix: str, external: str) -> str:
    raise NotImplementedError(__STUB_ERROR)


def decode_object_path(prefix: str, full_path: str) -> str:
    raise NotImplementedError(__STUB_ERROR)


def map_exception_to_dbus_error(exc: type[Exception],
                                dbus_error_name: str, /) -> None:
    ...  # We want to be able to generate docs without module


def add_exception_mapping(exc: type[Exception], /) -> None:
    ...  # We want to be able to generate docs without module


def is_interface_name_valid(string_to_check: str, /) -> bool:
    raise NotImplementedError(__STUB_ERROR)


def is_service_name_valid(string_to_check: str, /) -> bool:
    raise NotImplementedError(__STUB_ERROR)


def is_member_name_valid(string_to_check: str, /) -> bool:
    raise NotImplementedError(__STUB_ERROR)


def is_object_path_valid(string_to_check: str, /) -> bool:
    raise NotImplementedError(__STUB_ERROR)


class SdBusBaseError(Exception):
    ...


class SdBusUnmappedMessageError(SdBusBaseError):
    ...


class SdBusLibraryError(SdBusBaseError):
    ...


class SdBusRequestNameError(SdBusBaseError):
    ...


class SdBusRequestNameInQueueError(SdBusRequestNameError):
    ...


class SdBusRequestNameExistsError(SdBusRequestNameError):
    ...


class SdBusRequestNameAlreadyOwnerError(SdBusRequestNameError):
    ...


DBUS_ERROR_TO_EXCEPTION: dict[str, type[Exception]] = {}

EXCEPTION_TO_DBUS_ERROR: dict[type[Exception], str] = {}

DbusDeprecatedFlag: int = 0
DbusHiddenFlag: int = 0
DbusUnprivilegedFlag: int = 0
DbusNoReplyFlag: int = 0
DbusPropertyConstFlag: int = 0
DbusPropertyEmitsChangeFlag: int = 0
DbusPropertyEmitsInvalidationFlag: int = 0
DbusPropertyExplicitFlag: int = 0
DbusSensitiveFlag: int = 0

NameAllowReplacementFlag: int = 0
NameReplaceExistingFlag: int = 0
NameQueueFlag: int = 0



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2024 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from typing import TYPE_CHECKING

from ..dbus_common_elements import DbusLocalObjectMeta, DbusRemoteObjectMeta
from ..dbus_proxy_async_interface_base import DbusInterfaceBaseAsync
from ..dbus_proxy_sync_interface_base import DbusInterfaceBase
from ..default_bus import get_default_bus

if TYPE_CHECKING:
    from typing import Optional, Union

    from ..sd_bus_internals import SdBus


def _inspect_dbus_path_proxy(
    obj: object,
    dbus_meta: DbusRemoteObjectMeta,
    bus: SdBus,
) -> str:
    if bus != dbus_meta.attached_bus:
        raise LookupError(
            f"D-Bus proxy {obj!r} at {dbus_meta.object_path!r} path "
            f"is not attached to bus {bus!r}"
        )

    return dbus_meta.object_path


def _inspect_dbus_path_local(
    obj: object,
    dbus_meta: DbusLocalObjectMeta,
    bus: SdBus,
) -> str:
    attached_bus = dbus_meta.attached_bus
    object_path = dbus_meta.serving_object_path
    if attached_bus is None or object_path is None:
        raise LookupError(
            f"Local D-Bus object {obj!r} is not exported to any D-Bus"
        )

    if bus != attached_bus:
        raise LookupError(
            f"Local D-Bus object {obj!r} at {dbus_meta.serving_object_path!r} "
            f"path is not attached to bus {bus!r}"
        )

    return object_path


def inspect_dbus_path(
    obj: Union[DbusInterfaceBase, DbusInterfaceBaseAsync],
    bus: Optional[SdBus] = None,
) -> str:
    """Return the D-Bus path of an object.

    If called on a D-Bus proxy returns path of the proxied object.

    If called on a local D-Bus object returns the exported D-Bus path.
    If object is not exported raises ``LookupError``.

    If called on an object that is unrelated to D-Bus raises ``TypeError``.

    The object's path is inspected in the context of the given bus and if the
    object is attached to a different bus the ``LookupError`` will be raised.
    If the bus argument is not given or is ``None`` the default bus will be
    checked against.

    :param obj:
        Object to inspect.
    :param bus:
        Bus to inspect against.
        If not given or is ``None`` the default bus will be used.
    :returns:
        D-Bus path of the object.

    *New in version 0.13.0.*
    """
    if bus is None:
        bus = get_default_bus()

    if isinstance(obj, DbusInterfaceBase):
        return _inspect_dbus_path_proxy(obj, obj._dbus, bus)
    elif isinstance(obj, DbusInterfaceBaseAsync):
        dbus_meta = obj._dbus
        if isinstance(dbus_meta, DbusRemoteObjectMeta):
            return _inspect_dbus_path_proxy(obj, dbus_meta, bus)
        else:
            return _inspect_dbus_path_local(obj, dbus_meta, bus)
    else:
        raise TypeError(f"Expected D-Bus object got {obj!r}")


def inspect_dbus_bus(
    obj: Union[DbusInterfaceBase, DbusInterfaceBaseAsync]
) -> Optional[SdBus]:
    """Return D-Bus bus used by the object.

    If called on D-Bus proxies or exported local D-Bus objects returns
    bus object.

    If called on local D-Bus objects that had not been exported returns None.

    If called on an object that is unrelated to D-Bus raises ``TypeError``.

    :param obj:
        Object to inspect.
    :returns:
        D-Bus bus object.

    *New in version 0.14.1.*
    """
    if isinstance(obj, (DbusInterfaceBase, DbusInterfaceBaseAsync)):
        return obj._dbus.attached_bus
    else:
        raise TypeError(f"Expected D-Bus object got {obj!r}")


__all__ = (
    'inspect_dbus_bus',
    "inspect_dbus_path",
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from inspect import getfullargspec
from typing import TYPE_CHECKING, Generic, TypeVar

from .dbus_common_funcs import (
    _is_property_flags_correct,
    snake_case_to_camel_case,
)
from .default_bus import get_default_bus
from .sd_bus_internals import is_interface_name_valid, is_member_name_valid

if TYPE_CHECKING:
    from asyncio import Task
    from collections.abc import Callable, Sequence
    from types import FunctionType
    from typing import Any, Optional

    SelfMeta = TypeVar('SelfMeta', bound="DbusInterfaceMetaCommon")

    from .sd_bus_internals import SdBus, SdBusInterface

T = TypeVar('T')


class DbusMemberCommon:
    interface_name: str
    serving_enabled: bool


class DbusMemberAsync(DbusMemberCommon):
    ...


class DbusMemberSync(DbusMemberCommon):
    ...


class DbusInterfaceMetaCommon(type):
    def __new__(cls: type[SelfMeta], name: str,
                bases: tuple[type, ...],
                namespace: dict[str, Any],
                interface_name: Optional[str] = None,
                serving_enabled: bool = True,
                ) -> SelfMeta:
        if interface_name is not None:
            try:
                assert is_interface_name_valid(interface_name), (
                    f"Invalid interface name: \"{interface_name}\"; "
                    'Interface names must be composed of 2 or more elements '
                    'separated by a dot \'.\' character. All elements must '
                    'contain at least one character, consist of ASCII '
                    'characters, first character must not be digit and '
                    'length must not exceed 255 characters.'
                )
            except NotImplementedError:
                ...

        for attr_name, attr in namespace.items():
            if not isinstance(attr, DbusMemberCommon):
                continue

            # TODO: Fix async metaclass copying all methods
            if hasattr(attr, "interface_name"):
                continue

            if interface_name is None:
                raise TypeError(
                    f"Defined D-Bus element {attr_name!r} without "
                    f"interface name in the class {name!r}."
                )

            attr.interface_name = interface_name
            attr.serving_enabled = serving_enabled

        new_cls = super().__new__(cls, name, bases, namespace)

        return new_cls


MEMBER_NAME_REQUIREMENTS = (
    'Member name must only contain ASCII characters, '
    'cannot start with digit, '
    'must not contain dot \'.\' and be between 1 '
    'and 255 characters in length.'
)


class DbusMethodCommon(DbusMemberCommon):

    def __init__(
            self,
            original_method: FunctionType,
            method_name: Optional[str],
            input_signature: str,
            input_args_names: Optional[Sequence[str]],
            result_signature: str,
            result_args_names: Optional[Sequence[str]],
            flags: int):

        assert not isinstance(input_args_names, str), (
            "Passed a string as input args"
            " names. Did you forget to put"
            " it in to a tuple ('string', ) ?")

        if method_name is None:
            method_name = snake_case_to_camel_case(original_method.__name__)

        try:
            assert is_member_name_valid(method_name), (
                f"Invalid method name: \"{method_name}\"; "
                f"{MEMBER_NAME_REQUIREMENTS}"
            )
        except NotImplementedError:
            ...

        super().__init__()
        self.original_method = original_method
        self.args_spec = getfullargspec(original_method)
        self.args_names = self.args_spec.args[1:]  # 1: because of self
        self.num_of_args = len(self.args_names)
        self.args_defaults = (
            self.args_spec.defaults
            if self.args_spec.defaults is not None
            else ())
        self.default_args_start_at = self.num_of_args - len(self.args_defaults)

        self.method_name = method_name
        self.input_signature = input_signature
        self.input_args_names: Sequence[str] = ()
        if input_args_names is not None:
            assert not any(' ' in x for x in input_args_names), (
                "Can't have spaces in argument input names"
                f"Args: {input_args_names}")

            self.input_args_names = input_args_names
        elif result_args_names is not None:
            self.input_args_names = self.args_names

        self.result_signature = result_signature
        self.result_args_names: Sequence[str] = ()
        if result_args_names is not None:
            assert not any(' ' in x for x in result_args_names), (
                "Can't have spaces in argument result names."
                f"Args: {result_args_names}")

            self.result_args_names = result_args_names

        self.flags = flags

        self.__doc__ = original_method.__doc__

    def _rebuild_args(
            self,
            function: FunctionType,
            *args: Any,
            **kwargs: dict[str, Any]) -> list[Any]:
        # 3 types of arguments
        # *args - should be passed directly
        # **kwargs - should be put in a proper order
        # defaults - should be retrieved and put in proper order

        # Strategy:
        # Iterate over arg names
        # Use:
        # 1. Arg
        # 2. Kwarg
        # 3. Default

        # a, b, c, d, e
        #       ^ defaults start here
        # 5 - 3 = [2]
        # ^ total args
        #     ^ number of default args
        # First arg that supports default is
        # (total args - number of default args)
        passed_args_iter = iter(args)
        default_args_iter = iter(self.args_defaults)

        new_args_list: list[Any] = []

        for i, a_name in enumerate(self.args_spec.args[1:]):
            try:
                next_arg = next(passed_args_iter)
            except StopIteration:
                next_arg = None

            if i >= self.default_args_start_at:
                next_default_arg = next(default_args_iter)
            else:
                next_default_arg = None

            next_kwarg = kwargs.get(a_name)

            if next_arg is not None:
                new_args_list.append(next_arg)
            elif next_kwarg is not None:
                new_args_list.append(next_kwarg)
            elif next_default_arg is not None:
                new_args_list.append(next_default_arg)
            else:
                raise TypeError('Could not flatten the args')

        return new_args_list


class DbusPropertyCommon(DbusMemberCommon):
    def __init__(self,
                 property_name: Optional[str],
                 property_signature: str,
                 flags: int,
                 original_method: FunctionType):
        if property_name is None:
            property_name = snake_case_to_camel_case(original_method.__name__)

        try:
            assert is_member_name_valid(property_name), (
                f"Invalid property name: \"{property_name}\"; "
                f"{MEMBER_NAME_REQUIREMENTS}"
            )
        except NotImplementedError:
            ...

        assert _is_property_flags_correct(flags), (
            'Incorrect number of Property flags. '
            'Only one of DbusPropertyConstFlag, DbusPropertyEmitsChangeFlag, '
            'DbusPropertyEmitsInvalidationFlag or DbusPropertyExplicitFlag '
            'is allowed.'
        )

        super().__init__()
        self.property_name: str = property_name
        self.property_signature = property_signature
        self.flags = flags


class DbusSignalCommon(DbusMemberCommon):
    def __init__(self,
                 signal_name: Optional[str],
                 signal_signature: str,
                 args_names: Sequence[str],
                 flags: int,
                 original_method: FunctionType):
        if signal_name is None:
            signal_name = snake_case_to_camel_case(original_method.__name__)

        try:
            assert is_member_name_valid(signal_name), (
                f"Invalid signal name: \"{signal_name}\"; "
                f"{MEMBER_NAME_REQUIREMENTS}"
            )
        except NotImplementedError:
            ...

        super().__init__()
        self.signal_name = signal_name
        self.signal_signature = signal_signature
        self.args_names = args_names
        self.flags = flags

        self.__doc__ = original_method.__doc__
        self.__annotations__ = original_method.__annotations__


class DbusBoundAsync:
    ...


class DbusBoundSync:
    ...


class DbusMethodOverride(Generic[T]):
    def __init__(self, override_method: T):
        self.override_method = override_method


class DbusPropertyOverride(Generic[T]):
    def __init__(self, getter_override: Callable[[Any], T]):
        self.getter_override = getter_override
        self.setter_override: Optional[Callable[[Any, T], None]] = None
        self.is_setter_public = True

    def setter(self, new_setter: Optional[Callable[[Any, T], None]]) -> None:
        self.setter_override = new_setter

    def setter_private(
        self,
        new_setter: Optional[Callable[[Any, T], None]],
    ) -> None:
        self.setter_override = new_setter
        self.is_setter_public = False


class DbusRemoteObjectMeta:
    def __init__(
        self,
        service_name: str,
        object_path: str,
        bus: Optional[SdBus] = None,
    ):
        self.service_name = service_name
        self.object_path = object_path
        self.attached_bus = (
            bus if bus is not None
            else get_default_bus()
        )


class DbusLocalObjectMeta:
    def __init__(self) -> None:
        self.activated_interfaces: list[SdBusInterface] = []
        self.serving_object_path: Optional[str] = None
        self.attached_bus: Optional[SdBus] = None
        self._tasks: Optional[set[Task[None]]] = None

    @property
    def tasks(self) -> set[Task[None]]:
        tasks_set = self._tasks
        if tasks_set is None:
            tasks_set = set()
            self._tasks = tasks_set

        return tasks_set


class DbusClassMeta:
    def __init__(
        self,
        interface_name: str,
        serving_enabled: bool,
    ) -> None:
        self.interface_name = interface_name
        self.serving_enabled = serving_enabled
        self.dbus_member_to_python_attr: dict[str, str] = {}
        self.python_attr_to_dbus_member: dict[str, str] = {}



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2024 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from typing import TYPE_CHECKING

from .dbus_common_funcs import _parse_properties_vardict
from .dbus_proxy_async_interface_base import DbusInterfaceBaseAsync
from .dbus_proxy_async_method import dbus_method_async
from .dbus_proxy_async_signal import dbus_signal_async

if TYPE_CHECKING:
    from typing import Any, Literal

    DBUS_PROPERTIES_CHANGED_TYPING = (
        tuple[
            str,
            dict[str, tuple[str, Any]],
            list[str],
        ]
    )


class DbusPeerInterfaceAsync(
    DbusInterfaceBaseAsync,
    interface_name='org.freedesktop.DBus.Peer',
    serving_enabled=False,
):

    @dbus_method_async(method_name='Ping')
    async def dbus_ping(self) -> None:
        raise NotImplementedError

    @dbus_method_async(method_name='GetMachineId')
    async def dbus_machine_id(self) -> str:
        raise NotImplementedError


class DbusIntrospectableAsync(
    DbusInterfaceBaseAsync,
    interface_name='org.freedesktop.DBus.Introspectable',
    serving_enabled=False,
):

    @dbus_method_async(method_name='Introspect')
    async def dbus_introspect(self) -> str:
        raise NotImplementedError


class DbusPropertiesInterfaceAsync(
    DbusInterfaceBaseAsync,
    interface_name='org.freedesktop.DBus.Properties',
    serving_enabled=False,
):

    @dbus_signal_async('sa{sv}as')
    def properties_changed(self) -> DBUS_PROPERTIES_CHANGED_TYPING:
        raise NotImplementedError

    @dbus_method_async('s', 'a{sv}', method_name='GetAll')
    async def _properties_get_all(
            self, interface_name: str) -> dict[str, tuple[str, Any]]:
        raise NotImplementedError

    async def properties_get_all_dict(
            self,
            on_unknown_member: Literal['error', 'ignore', 'reuse'] = 'error',
    ) -> dict[str, Any]:

        properties: dict[str, Any] = {}

        for interface_name, meta in self._dbus_iter_interfaces_meta():
            if not meta.serving_enabled:
                continue

            dbus_properties_data = await self._properties_get_all(
                interface_name)

            properties.update(
                _parse_properties_vardict(
                    meta.dbus_member_to_python_attr,
                    dbus_properties_data,
                    on_unknown_member,
                )
            )

        return properties


class DbusInterfaceCommonAsync(
        DbusPropertiesInterfaceAsync,
        DbusIntrospectableAsync,
        DbusPeerInterfaceAsync):
    ...



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from gc import collect
from unittest import TestCase
from unittest import main as unittest_main

from sdbus import DbusInterfaceCommon, dbus_method, dbus_property

from .common_test_util import skip_if_no_name_validations


class GoodDbusInterface(
    DbusInterfaceCommon,
    interface_name="org.example.test",
):
    @dbus_method()
    def test_method(self) -> None:
        raise NotImplementedError

    @dbus_property("s")
    def test_property(self) -> str:
        return "test"


class TestBadDbusClass(TestCase):
    def test_method_name_override(self) -> None:
        with self.subTest("Method override"), self.assertRaises(ValueError):

            class BadMethodOverrideClass(GoodDbusInterface):
                def test_method(self) -> None:
                    return

        with self.subTest("D-Bus method override"), self.assertRaises(
            ValueError
        ):

            class BadDbusMethodOverrideClass(GoodDbusInterface):
                @dbus_method()
                def test_method(self) -> None:
                    return

        with self.subTest("Property override"), self.assertRaises(ValueError):

            class BadPropertyOverrideClass(GoodDbusInterface):
                def test_property(self) -> str:  # type: ignore
                    return "override"

        with self.subTest("D-Bus property override"), self.assertRaises(
            ValueError
        ):

            class BadDbusPropertyOverrideClass(GoodDbusInterface):
                @dbus_property("s")
                def test_property(self) -> str:
                    return "override"

        with self.subTest("Good new method"):

            class GoodSubclass(GoodDbusInterface):
                def new_method(self) -> int:
                    return 1

    def test_interface_collision(self) -> None:
        with self.subTest("No collision"):
            class NonInterface(GoodDbusInterface):
                def do_work(self) -> None:
                    ...

        with self.subTest("Collision"), self.assertRaises(ValueError):
            class NewExampleInterface(
                DbusInterfaceCommon,
                interface_name="org.example.test",
            ):
                ...

    def test_bad_class_names(self) -> None:
        skip_if_no_name_validations()

        with self.assertRaisesRegex(AssertionError, "^Invalid interface name"):

            class BadInterfaceName(
                DbusInterfaceCommon,
                interface_name="0.test",
            ):
                ...

        with self.assertRaisesRegex(
            AssertionError,
            "^Invalid method name",
        ):

            class BadMethodName(
                DbusInterfaceCommon,
                interface_name="org.example",
            ):
                @dbus_method(
                    result_signature="s",
                    method_name="🤫",
                )
                def test(self) -> str:
                    return "test"

        with self.assertRaisesRegex(
            AssertionError,
            "^Invalid property name",
        ):

            class BadPropertyName(
                DbusInterfaceCommon,
                interface_name="org.example",
            ):
                @dbus_property(
                    property_signature="s",
                    property_name="🤫",
                )
                def test(self) -> str:
                    return "test"

    def test_dbus_elements_without_interface_name(self) -> None:
        with self.assertRaisesRegex(TypeError, "without interface name"):

            class NoInterfaceName(DbusInterfaceCommon):
                @dbus_method()
                def example(self) -> None:
                    ...

    def test_shared_parent_class(self) -> None:
        class One(GoodDbusInterface):
            ...

        class Two(GoodDbusInterface):
            ...

        class Shared(One, Two):
            ...

    def test_combined_collision(self) -> None:

        class One(
            DbusInterfaceCommon,
            interface_name="org.example.foo",
        ):
            @dbus_method()
            def example(self) -> None:
                ...

        class Two(
            DbusInterfaceCommon,
            interface_name="org.example.bar",
        ):
            @dbus_method()
            def example(self) -> None:
                ...

        with self.assertRaisesRegex(ValueError, "collision"):
            class Combined(One, Two):
                ...

    def test_class_cleanup(self) -> None:
        class One(
            DbusInterfaceCommon,
            interface_name="org.example.foo1",
        ):
            ...

        with self.assertRaises(ValueError):
            class Two(
                DbusInterfaceCommon,
                interface_name="org.example.foo1",
            ):
                ...

        del One
        collect()  # Let weak refs be processed

        class After(
            DbusInterfaceCommon,
            interface_name="org.example.foo1",
        ):
            ...


if __name__ == "__main__":
    unittest_main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from gc import collect
from unittest import TestCase
from unittest import main as unittest_main

from sdbus.dbus_common_funcs import PROPERTY_FLAGS_MASK, count_bits

from sdbus import (
    DbusDeprecatedFlag,
    DbusInterfaceCommonAsync,
    DbusPropertyConstFlag,
    DbusPropertyEmitsChangeFlag,
    dbus_method_async,
    dbus_method_async_override,
    dbus_property_async,
    dbus_signal_async,
)

from .common_test_util import skip_if_no_asserts, skip_if_no_name_validations


class TestInterface(
    DbusInterfaceCommonAsync,
    interface_name="org.example.good",
):
    @dbus_method_async(result_signature="i")
    async def test_int(self) -> int:
        return 1


class TestBadAsyncDbusClass(TestCase):
    def test_name_validations(self) -> None:
        skip_if_no_name_validations()

        with self.assertRaisesRegex(
            AssertionError,
            "^Invalid interface name",
        ):

            class BadInterfaceName(
                DbusInterfaceCommonAsync,
                interface_name="0.test",
            ):
                ...

        with self.assertRaisesRegex(
            AssertionError,
            "^Invalid method name",
        ):

            class BadMethodName(
                DbusInterfaceCommonAsync,
                interface_name="org.example",
            ):
                @dbus_method_async(
                    result_signature="s",
                    method_name="🤫",
                )
                async def test(self) -> str:
                    return "test"

        with self.assertRaisesRegex(
            AssertionError,
            "^Invalid property name",
        ):

            class BadPropertyName(
                DbusInterfaceCommonAsync,
                interface_name="org.example",
            ):
                @dbus_property_async(
                    property_signature="s",
                    property_name="🤫",
                )
                def test(self) -> str:
                    return "test"

        with self.assertRaisesRegex(
            AssertionError,
            "^Invalid signal name",
        ):

            class BadSignalName(
                DbusInterfaceCommonAsync,
                interface_name="org.example",
            ):
                @dbus_signal_async(
                    signal_signature="s",
                    signal_name="🤫",
                )
                def test(self) -> str:
                    raise NotImplementedError

    def test_property_flags(self) -> None:
        self.assertEqual(0, PROPERTY_FLAGS_MASK & DbusDeprecatedFlag)
        self.assertEqual(
            1,
            count_bits(
                PROPERTY_FLAGS_MASK
                & (DbusDeprecatedFlag | DbusPropertyEmitsChangeFlag)
            ),
        )
        self.assertEqual(
            2,
            count_bits(
                PROPERTY_FLAGS_MASK
                & (
                    DbusDeprecatedFlag
                    | DbusPropertyConstFlag
                    | DbusPropertyEmitsChangeFlag
                )
            ),
        )

        with self.subTest("Test incorrect flags"), self.assertRaisesRegex(
            AssertionError,
            "^Incorrect number of Property flags",
        ):
            skip_if_no_asserts()

            class InvalidPropertiesFlags(
                DbusInterfaceCommonAsync, interface_name="org.test.invalidprop"
            ):
                @dbus_property_async(
                    "s",
                    flags=DbusPropertyConstFlag | DbusPropertyEmitsChangeFlag,
                )
                def test_constant(self) -> str:
                    return "a"

        with self.subTest("Valid properties flags"):

            class ValidPropertiesFlags(
                DbusInterfaceCommonAsync, interface_name="org.test.validprop"
            ):
                @dbus_property_async(
                    "s",
                    flags=DbusDeprecatedFlag | DbusPropertyEmitsChangeFlag,
                )
                def test_constant(self) -> str:
                    return "a"

    def test_bad_subclass(self) -> None:
        with self.assertRaises(ValueError):

            class TestInheritence(TestInterface):
                async def test_int(self) -> int:
                    return 2

        with self.assertRaises(ValueError):

            class TestInheritence2(TestInterface):
                @dbus_method_async_override()
                async def test_unrelated(self) -> int:
                    return 2

    def test_dbus_elements_without_interface_name(self) -> None:
        with self.assertRaisesRegex(TypeError, "without interface name"):

            class NoInterfaceName(DbusInterfaceCommonAsync):
                @dbus_method_async()
                async def example(self) -> None:
                    ...

    def test_dbus_elements_without_interface_name_subclass(self) -> None:
        with self.assertRaisesRegex(TypeError, "without interface name"):

            class NoInterfaceName(TestInterface):
                @dbus_method_async()
                async def example(self) -> None:
                    ...

    def test_shared_parent_class(self) -> None:
        class One(TestInterface):
            ...

        class Two(TestInterface):
            ...

        class Shared(One, Two):
            ...

    def test_combined_collision(self) -> None:

        class One(
            DbusInterfaceCommonAsync,
            interface_name="org.example.foo",
        ):
            @dbus_method_async()
            async def example(self) -> None:
                ...

        class Two(
            DbusInterfaceCommonAsync,
            interface_name="org.example.bar",
        ):
            @dbus_method_async()
            async def example(self) -> None:
                ...

        with self.assertRaisesRegex(ValueError, "collision"):
            class Combined(One, Two):
                ...

    def test_class_cleanup(self) -> None:
        class One(
            DbusInterfaceCommonAsync,
            interface_name="org.example.foo1",
        ):
            ...

        with self.assertRaises(ValueError):
            class Two(
                DbusInterfaceCommonAsync,
                interface_name="org.example.foo1",
            ):
                ...

        del One
        collect()  # Let weak refs be processed

        class After(
            DbusInterfaceCommonAsync,
            interface_name="org.example.foo1",
        ):
            ...


if __name__ == "__main__":
    unittest_main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import get_running_loop
from contextvars import ContextVar, copy_context
from inspect import iscoroutinefunction
from types import FunctionType
from typing import TYPE_CHECKING, cast, overload
from weakref import ref as weak_ref

from .dbus_common_elements import (
    DbusBoundAsync,
    DbusLocalObjectMeta,
    DbusMemberAsync,
    DbusMethodCommon,
    DbusMethodOverride,
    DbusRemoteObjectMeta,
)
from .dbus_exceptions import DbusFailedError
from .sd_bus_internals import EXCEPTION_TO_DBUS_ERROR, DbusNoReplyFlag

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Optional, TypeVar, Union

    from .dbus_proxy_async_interface_base import DbusInterfaceBaseAsync
    from .sd_bus_internals import SdBusMessage

    T = TypeVar('T')
else:
    T = None

CURRENT_MESSAGE: ContextVar[SdBusMessage] = ContextVar('CURRENT_MESSAGE')


def get_current_message() -> SdBusMessage:
    return CURRENT_MESSAGE.get()


class DbusMethodAsync(DbusMethodCommon, DbusMemberAsync):

    @overload
    def __get__(
        self,
        obj: None,
        obj_class: type[DbusInterfaceBaseAsync],
    ) -> DbusMethodAsync:
        ...

    @overload
    def __get__(
        self,
        obj: DbusInterfaceBaseAsync,
        obj_class: type[DbusInterfaceBaseAsync],
    ) -> Callable[..., Any]:
        ...

    def __get__(
        self,
        obj: Optional[DbusInterfaceBaseAsync],
        obj_class: Optional[type[DbusInterfaceBaseAsync]] = None,
    ) -> Union[Callable[..., Any], DbusMethodAsync]:
        if obj is not None:
            dbus_meta = obj._dbus
            if isinstance(dbus_meta, DbusRemoteObjectMeta):
                return DbusProxyMethodAsync(self, dbus_meta)
            else:
                return DbusLocalMethodAsync(self, obj)
        else:
            return self


class DbusBoundMethodAsyncBase(DbusBoundAsync):

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError


class DbusProxyMethodAsync(DbusBoundMethodAsyncBase):
    def __init__(
        self,
        dbus_method: DbusMethodAsync,
        proxy_meta: DbusRemoteObjectMeta,
    ):
        self.dbus_method = dbus_method
        self.proxy_meta = proxy_meta

        self.__doc__ = dbus_method.__doc__

    async def _dbus_async_call(self, call_message: SdBusMessage) -> Any:
        bus = self.proxy_meta.attached_bus
        reply_message = await bus.call_async(call_message)
        return reply_message.get_contents()

    @staticmethod
    async def _no_reply() -> None:
        return None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        bus = self.proxy_meta.attached_bus
        dbus_method = self.dbus_method

        new_call_message = bus.new_method_call_message(
            self.proxy_meta.service_name,
            self.proxy_meta.object_path,
            dbus_method.interface_name,
            dbus_method.method_name,
        )

        if len(args) == dbus_method.num_of_args:
            assert not kwargs, (
                "Passed more arguments than method supports"
                f"Extra args: {kwargs}")
            rebuilt_args: Sequence[Any] = args
        else:
            rebuilt_args = dbus_method._rebuild_args(
                dbus_method.original_method,
                *args,
                **kwargs)

        if rebuilt_args:
            new_call_message.append_data(
                dbus_method.input_signature, *rebuilt_args)

        if dbus_method.flags & DbusNoReplyFlag:
            new_call_message.expect_reply = False
            new_call_message.send()
            return self._no_reply()

        return self._dbus_async_call(new_call_message)


class DbusLocalMethodAsync(DbusBoundMethodAsyncBase):
    def __init__(
        self,
        dbus_method: DbusMethodAsync,
        local_object: DbusInterfaceBaseAsync,
    ):
        self.dbus_method = dbus_method
        self.local_object_ref = weak_ref(local_object)

        self.__doc__ = dbus_method.__doc__

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        local_object = self.local_object_ref()
        if local_object is None:
            raise RuntimeError("Local object no longer exists!")

        return self.dbus_method.original_method(local_object, *args, **kwargs)

    async def _dbus_reply_call_method(
        self,
        request_message: SdBusMessage,
        local_object: DbusInterfaceBaseAsync,
    ) -> Any:

        local_method = self.dbus_method.original_method.__get__(
            local_object, None)

        CURRENT_MESSAGE.set(request_message)

        return await local_method(*request_message.parse_to_tuple())

    def _dbus_reply_call(
        self,
        request_message: SdBusMessage
    ) -> None:
        local_object = self.local_object_ref()
        if local_object is None:
            raise RuntimeError("Local object no longer exists!")
        local_meta = local_object._dbus
        if not isinstance(local_meta, DbusLocalObjectMeta):
            raise RuntimeError("D-Bus object is a remote proxy!")

        loop = get_running_loop()
        reply_task = loop.create_task(
            self._dbus_reply_call_async(local_object, request_message)
        )
        tasks_set = local_meta.tasks
        tasks_set.add(reply_task)
        reply_task.add_done_callback(tasks_set.discard)

    async def _dbus_reply_call_async(
        self,
        local_object: DbusInterfaceBaseAsync,
        request_message: SdBusMessage
    ) -> None:
        call_context = copy_context()
        try:
            reply_data = await call_context.run(
                self._dbus_reply_call_method,
                request_message,
                local_object,
            )
        except Exception as e:
            if not request_message.expect_reply:
                return

            dbus_error = EXCEPTION_TO_DBUS_ERROR.get(type(e))
            if dbus_error is None:
                dbus_error = DbusFailedError.dbus_error_name

            error_message = request_message.create_error_reply(
                dbus_error,
                str(e.args[0]) if e.args else "",
            )
            error_message.send()
            return

        if not request_message.expect_reply:
            return

        reply_message = request_message.create_reply()

        if isinstance(reply_data, tuple):
            try:
                reply_message.append_data(
                    self.dbus_method.result_signature, *reply_data)
            except TypeError:
                # In case of single struct result type
                # We can't figure out if return is multiple values
                # or a tuple
                reply_message.append_data(
                    self.dbus_method.result_signature, reply_data)
        elif reply_data is not None:
            reply_message.append_data(
                self.dbus_method.result_signature, reply_data)

        reply_message.send()


def dbus_method_async(
    input_signature: str = "",
    result_signature: str = "",
    flags: int = 0,
    result_args_names: Optional[Sequence[str]] = None,
    input_args_names: Optional[Sequence[str]] = None,
    method_name: Optional[str] = None,
) -> Callable[[T], T]:

    assert not isinstance(input_signature, FunctionType), (
        "Passed function to decorator directly. "
        "Did you forget () round brackets?"
    )

    def dbus_method_decorator(original_method: T) -> T:
        assert isinstance(original_method, FunctionType)
        assert iscoroutinefunction(original_method), (
            "Expected coroutine function. ",
            "Maybe you forgot 'async' keyword?",
        )
        new_wrapper = DbusMethodAsync(
            original_method=original_method,
            method_name=method_name,
            input_signature=input_signature,
            result_signature=result_signature,
            result_args_names=result_args_names,
            input_args_names=input_args_names,
            flags=flags,
        )

        return cast(T, new_wrapper)

    return dbus_method_decorator


def dbus_method_async_override() -> Callable[[T], T]:

    def new_decorator(
            new_function: T) -> T:
        return cast(T, DbusMethodOverride(new_function))

    return new_decorator



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from .dbus_exceptions import (
    DbusAccessDeniedError,
    DbusAddressInUseError,
    DbusAuthFailedError,
    DbusBadAddressError,
    DbusDisconnectedError,
    DbusFailedError,
    DbusFileExistsError,
    DbusFileNotFoundError,
    DbusInconsistentMessageError,
    DbusInteractiveAuthorizationRequiredError,
    DbusInvalidArgsError,
    DbusInvalidFileContentError,
    DbusInvalidSignatureError,
    DbusIOError,
    DbusLimitsExceededError,
    DbusMatchRuleInvalidError,
    DbusMatchRuleNotFound,
    DbusNameHasNoOwnerError,
    DbusNoMemoryError,
    DbusNoNetworkError,
    DbusNoReplyError,
    DbusNoServerError,
    DbusNotSupportedError,
    DbusPropertyReadOnlyError,
    DbusServiceUnknownError,
    DbusTimeoutError,
    DbusUnixProcessIdUnknownError,
    DbusUnknownInterfaceError,
    DbusUnknownMethodError,
    DbusUnknownObjectError,
    DbusUnknownPropertyError,
)
from .dbus_proxy_async_interfaces import DbusInterfaceCommonAsync
from .dbus_proxy_async_method import (
    dbus_method_async,
    dbus_method_async_override,
    get_current_message,
)
from .dbus_proxy_async_object_manager import DbusObjectManagerInterfaceAsync
from .dbus_proxy_async_property import (
    dbus_property_async,
    dbus_property_async_override,
)
from .dbus_proxy_async_signal import dbus_signal_async
from .dbus_proxy_sync_interfaces import (
    DbusInterfaceCommon,
    DbusObjectManagerInterface,
)
from .dbus_proxy_sync_method import dbus_method
from .dbus_proxy_sync_property import dbus_property
from .default_bus import (
    get_default_bus,
    request_default_bus_name,
    request_default_bus_name_async,
    set_context_default_bus,
    set_default_bus,
)
from .sd_bus_internals import (
    DbusDeprecatedFlag,
    DbusHiddenFlag,
    DbusNoReplyFlag,
    DbusPropertyConstFlag,
    DbusPropertyEmitsChangeFlag,
    DbusPropertyEmitsInvalidationFlag,
    DbusPropertyExplicitFlag,
    DbusSensitiveFlag,
    DbusUnprivilegedFlag,
    SdBus,
    SdBusBaseError,
    SdBusLibraryError,
    SdBusUnmappedMessageError,
    decode_object_path,
    encode_object_path,
    map_exception_to_dbus_error,
    sd_bus_open,
    sd_bus_open_system,
    sd_bus_open_system_machine,
    sd_bus_open_system_remote,
    sd_bus_open_user,
    sd_bus_open_user_machine,
)

__all__ = (
    'DbusAccessDeniedError', 'DbusAddressInUseError',
    'DbusAuthFailedError', 'DbusBadAddressError',
    'DbusDisconnectedError', 'DbusFailedError',
    'DbusFileExistsError', 'DbusFileNotFoundError',
    'DbusInconsistentMessageError',
    'DbusInteractiveAuthorizationRequiredError',
    'DbusInvalidArgsError',
    'DbusInvalidFileContentError',
    'DbusInvalidSignatureError', 'DbusIOError',
    'DbusLimitsExceededError',
    'DbusMatchRuleInvalidError', 'DbusMatchRuleNotFound',
    'DbusNameHasNoOwnerError', 'DbusNoMemoryError',
    'DbusNoNetworkError', 'DbusNoReplyError',
    'DbusNoServerError', 'DbusNotSupportedError',
    'DbusPropertyReadOnlyError',
    'DbusServiceUnknownError', 'DbusTimeoutError',
    'DbusUnixProcessIdUnknownError',
    'DbusUnknownInterfaceError',
    'DbusUnknownMethodError', 'DbusUnknownObjectError',
    'DbusUnknownPropertyError',

    'DbusInterfaceCommonAsync',
    'DbusObjectManagerInterfaceAsync',

    'dbus_method_async',
    'dbus_method_async_override',

    'dbus_property_async',
    'dbus_property_async_override',
    'get_current_message',

    'dbus_signal_async',

    'DbusInterfaceCommon',
    'DbusObjectManagerInterface',

    'dbus_method',

    'dbus_property',

    "get_default_bus",
    "request_default_bus_name",
    "request_default_bus_name_async",
    "set_context_default_bus",
    "set_default_bus",

    'DbusDeprecatedFlag',
    'DbusHiddenFlag',
    'DbusNoReplyFlag',
    'DbusPropertyConstFlag',
    'DbusPropertyEmitsChangeFlag',
    'DbusPropertyEmitsInvalidationFlag',
    'DbusPropertyExplicitFlag',
    'DbusSensitiveFlag',
    'DbusUnprivilegedFlag',
    'SdBus',
    'SdBusBaseError',
    'SdBusLibraryError',
    'SdBusUnmappedMessageError',
    'decode_object_path',
    'encode_object_path',
    'map_exception_to_dbus_error',
    'sd_bus_open',
    'sd_bus_open_system',
    'sd_bus_open_system_machine',
    'sd_bus_open_system_remote',
    'sd_bus_open_user',
    'sd_bus_open_user_machine',
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from sdbus.exceptions import DbusUnknownObjectError
from sdbus.unittest import IsolatedDbusTestCase
from sdbus.utils import (
    parse_get_managed_objects,
    parse_interfaces_added,
    parse_interfaces_removed,
)

from sdbus import (
    DbusInterfaceCommonAsync,
    DbusObjectManagerInterfaceAsync,
    dbus_method_async,
    dbus_property_async,
)

HELLO_WORLD = 'Hello World!'
TEST_NUMBER = 1000


class ObjectManagerTestInterface(
    DbusObjectManagerInterfaceAsync,
    interface_name='org.test.objectmanager',
):
    @dbus_method_async(
        result_signature='s',
    )
    async def get_hello_world(self) -> str:
        return HELLO_WORLD


OBJECT_MANAGER_PATH = '/object_manager'
CONNECTION_NAME = 'org.example.test'

MANAGED_INTERFACE_NAME = 'org.test.testing'


class ManagedInterface(
    DbusInterfaceCommonAsync,
    interface_name=MANAGED_INTERFACE_NAME,
):

    @dbus_property_async('x')
    def test_int(self) -> int:
        return TEST_NUMBER


MANAGED_PATH = '/object_manager/test'


class TestObjectManager(IsolatedDbusTestCase):
    async def test_object_manager(self) -> None:
        await self.bus.request_name_async(CONNECTION_NAME, 0)

        object_manager = ObjectManagerTestInterface()
        object_manager.export_to_dbus(OBJECT_MANAGER_PATH)

        object_manager_connection = ObjectManagerTestInterface.new_proxy(
            CONNECTION_NAME, OBJECT_MANAGER_PATH)

        self.assertEqual(
            await object_manager_connection.get_hello_world(),
            HELLO_WORLD)

        managed_object = ManagedInterface()

        async with self.assertDbusSignalEmits(
            object_manager_connection.interfaces_added
        ) as added_interfaces_catch:
            object_manager.export_with_manager(MANAGED_PATH, managed_object)

        caught_added = added_interfaces_catch.output[0]

        added_path, added_attributes = caught_added

        self.assertEqual(added_path, MANAGED_PATH)

        self.assertEqual(
            added_attributes[
                MANAGED_INTERFACE_NAME][
                    'TestInt'][1],
            TEST_NUMBER,
        )

        with self.subTest("Test interfaces added parser"):
            parse_interfaces_added(ManagedInterface, caught_added)

        async with self.assertDbusSignalEmits(
            object_manager_connection.interfaces_removed
        ) as removed_interfaces_catch:
            object_manager.remove_managed_object(managed_object)

        path_removed, interfaces_removed = removed_interfaces_catch.output[0]

        self.assertEqual(path_removed, MANAGED_PATH)

        self.assertIn(MANAGED_INTERFACE_NAME, interfaces_removed)

    def test_expot_with_no_manager(self) -> None:
        object_manager = ObjectManagerTestInterface()

        managed_object = ManagedInterface()

        self.assertRaises(
            RuntimeError,
            object_manager.export_with_manager,
            MANAGED_PATH,
            managed_object,
        )

    async def test_parse_interfaces_added_removed(self) -> None:
        MANAGED_TWO_INTERFACE_NAME = MANAGED_INTERFACE_NAME + 'Two'

        class ManagedTwoInterface(
            ManagedInterface,
            interface_name=MANAGED_TWO_INTERFACE_NAME,
        ):

            @dbus_property_async('s')
            def test_str(self) -> str:
                return 'test'

        await self.bus.request_name_async(CONNECTION_NAME, 0)

        object_manager = DbusObjectManagerInterfaceAsync()
        object_manager.export_to_dbus(OBJECT_MANAGER_PATH)

        object_manager_connection = DbusObjectManagerInterfaceAsync.new_proxy(
            CONNECTION_NAME, OBJECT_MANAGER_PATH)

        managed_object = ManagedTwoInterface()

        async with self.assertDbusSignalEmits(
            object_manager_connection.interfaces_added
        ) as added_interfaces_catch:
            object_manager.export_with_manager(MANAGED_PATH, managed_object)

        caught_added = added_interfaces_catch.output[0]

        with self.subTest('Parse added class'):
            path, python_class, python_properties = (
                parse_interfaces_added(ManagedTwoInterface, caught_added)
            )

            self.assertEqual(path, MANAGED_PATH)
            self.assertEqual(python_class, ManagedTwoInterface)
            self.assertIn('test_str', python_properties)
            self.assertIn('test_int', python_properties)

        with self.subTest('Parse added object'):
            path, python_class, python_properties = (
                parse_interfaces_added(managed_object, caught_added)
            )

            self.assertEqual(path, MANAGED_PATH)
            self.assertEqual(python_class, ManagedTwoInterface)
            self.assertIn('test_str', python_properties)
            self.assertIn('test_int', python_properties)

        with self.subTest('Parse added iterable'):
            path, python_class, python_properties = (
                parse_interfaces_added(
                    (ManagedInterface, ManagedTwoInterface),
                    caught_added)
            )

            self.assertEqual(path, MANAGED_PATH)
            self.assertEqual(python_class, ManagedTwoInterface)
            self.assertIn('test_str', python_properties)
            self.assertIn('test_int', python_properties)

        with self.subTest('Parse added unknown'):
            with self.assertRaises(KeyError):
                path, python_class, python_properties = (
                    parse_interfaces_added(
                        ManagedInterface,
                        caught_added)
                )

            with self.assertRaises(KeyError):
                path, python_class, python_properties = (
                    parse_interfaces_added(
                        ManagedInterface,
                        caught_added,
                        on_unknown_interface='none',
                    )
                )

            path, python_class, python_properties = (
                parse_interfaces_added(
                    ManagedInterface,
                    caught_added,
                    on_unknown_interface='none',
                    on_unknown_member='reuse',
                )
            )

            self.assertEqual(path, MANAGED_PATH)
            self.assertIsNone(python_class)
            self.assertIn('TestStr', python_properties)
            self.assertIn('TestInt', python_properties)

        get_managed_data = (
            await object_manager_connection.get_managed_objects()
        )
        with self.subTest('Parse get managed objects class'):
            managed_dict = (
                parse_get_managed_objects(
                    ManagedTwoInterface,
                    get_managed_data,
                )
            )

            self.assertIn(MANAGED_PATH, managed_dict)
            managed_class, managed_properties = (
                managed_dict[MANAGED_PATH]
            )
            self.assertEqual(managed_class, ManagedTwoInterface)
            self.assertIn('test_str', managed_properties)
            self.assertIn('test_int', managed_properties)

        with self.subTest('Parse get managed objects object'):
            managed_dict = (
                parse_get_managed_objects(
                    managed_object,
                    get_managed_data,
                )
            )

            self.assertIn(MANAGED_PATH, managed_dict)
            managed_class, managed_properties = (
                managed_dict[MANAGED_PATH]
            )
            self.assertEqual(managed_class, ManagedTwoInterface)
            self.assertIn('test_str', managed_properties)
            self.assertIn('test_int', managed_properties)

        with self.subTest('Parse get managed objects iterable'):
            managed_dict = (
                parse_get_managed_objects(
                    (ManagedInterface, ManagedTwoInterface),
                    get_managed_data,
                )
            )

            self.assertIn(MANAGED_PATH, managed_dict)
            managed_class, managed_properties = (
                managed_dict[MANAGED_PATH]
            )
            self.assertEqual(managed_class, ManagedTwoInterface)
            self.assertIn('test_str', managed_properties)
            self.assertIn('test_int', managed_properties)

        with self.subTest('Parse get managed objects unknown'):
            with self.assertRaises(KeyError):
                managed_dict = (
                    parse_get_managed_objects(
                        ManagedInterface,
                        get_managed_data,
                    )
                )

            with self.assertRaises(KeyError):
                managed_dict = (
                    parse_get_managed_objects(
                        ManagedInterface,
                        get_managed_data,
                        on_unknown_interface='none',
                    )
                )

            managed_dict = (
                parse_get_managed_objects(
                    ManagedInterface,
                    get_managed_data,
                    on_unknown_interface='none',
                    on_unknown_member='reuse',
                )
            )
            path, python_class, python_properties = (
                parse_interfaces_added(
                    ManagedInterface,
                    caught_added,
                    on_unknown_interface='none',
                    on_unknown_member='reuse',
                )
            )

            self.assertIn(MANAGED_PATH, managed_dict)
            managed_class, managed_properties = (
                managed_dict[MANAGED_PATH]
            )
            self.assertIsNone(managed_class)
            self.assertIn('TestStr', managed_properties)
            self.assertIn('TestInt', managed_properties)

        async with self.assertDbusSignalEmits(
            object_manager_connection.interfaces_removed
        ) as removed_interfaces_catch:
            object_manager.remove_managed_object(managed_object)

        interfaces_removed_data = removed_interfaces_catch.output[0]

        with self.subTest('Parse removed class'):
            path, python_class = (
                parse_interfaces_removed(
                    ManagedTwoInterface,
                    interfaces_removed_data,
                )
            )

            self.assertEqual(path, MANAGED_PATH)
            self.assertEqual(python_class, ManagedTwoInterface)

        with self.subTest('Parse removed unknown'):
            with self.assertRaises(KeyError):
                path, python_class = (
                    parse_interfaces_removed(
                        ManagedInterface,
                        interfaces_removed_data,
                    )
                )

            path, python_class = (
                parse_interfaces_removed(
                    ManagedInterface,
                    interfaces_removed_data,
                    on_unknown_interface='none',
                )
            )

            self.assertEqual(path, MANAGED_PATH)
            self.assertIsNone(python_class)

    async def test_main_export_handle(self) -> None:
        await self.bus.request_name_async(CONNECTION_NAME, 0)

        object_manager = ObjectManagerTestInterface()

        object_manager_connection = ObjectManagerTestInterface.new_proxy(
            CONNECTION_NAME, OBJECT_MANAGER_PATH)

        with object_manager.export_to_dbus(OBJECT_MANAGER_PATH):
            self.assertIsInstance(
                await object_manager_connection.get_managed_objects(),
                dict,
            )

        with self.assertRaises(DbusUnknownObjectError):
            self.assertIsInstance(
                await object_manager_connection.get_managed_objects(),
                dict,
            )

    async def test_secondary_export_handle(self) -> None:
        await self.bus.request_name_async(CONNECTION_NAME, 0)

        object_manager = ObjectManagerTestInterface()

        object_manager_connection = ObjectManagerTestInterface.new_proxy(
            CONNECTION_NAME, OBJECT_MANAGER_PATH)
        object_manager.export_to_dbus(OBJECT_MANAGER_PATH)

        managed_object = ManagedInterface()
        managed_proxy = ManagedInterface.new_proxy(
            CONNECTION_NAME, MANAGED_PATH,
        )

        async with self.assertDbusSignalEmits(
            object_manager_connection.interfaces_added
        ) as added, self.assertDbusSignalEmits(
            object_manager_connection.interfaces_removed
        ) as removed, object_manager.export_with_manager(
            MANAGED_PATH, managed_object,
        ):
            self.assertEqual(
                await managed_proxy.test_int,
                TEST_NUMBER,
            )

        self.assertEqual(added.output[0][0], MANAGED_PATH)

        removed_path, removed_interfaces = removed.output[0]
        self.assertEqual(removed_path, MANAGED_PATH)
        self.assertIn(MANAGED_INTERFACE_NAME, removed_interfaces)

        with self.assertRaises(DbusUnknownObjectError):
            await managed_proxy.test_int



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import get_running_loop, sleep, wait_for
from unittest import main

from sdbus.exceptions import (
    SdBusLibraryError,
    SdBusRequestNameAlreadyOwnerError,
    SdBusRequestNameError,
    SdBusRequestNameExistsError,
    SdBusRequestNameInQueueError,
)
from sdbus.sd_bus_internals import NameAllowReplacementFlag, NameQueueFlag
from sdbus.unittest import IsolatedDbusTestCase
from sdbus_async.dbus_daemon import FreedesktopDbus

from sdbus import (
    request_default_bus_name,
    request_default_bus_name_async,
    sd_bus_open_user,
)

TEST_BUS_NAME = 'com.example.test'
TEST_BUS_NAME_regex_match = TEST_BUS_NAME.replace('.', r'\.')


class TestRequestNameLowLevel(IsolatedDbusTestCase):
    def test_request_name_exception_tree(self) -> None:
        # Test that SdBusRequestNameError is super class
        # of other request name exceptions
        self.assertTrue(
            issubclass(
                SdBusRequestNameAlreadyOwnerError,
                SdBusRequestNameError,
            )
        )
        self.assertTrue(
            issubclass(
                SdBusRequestNameExistsError,
                SdBusRequestNameError,
            )
        )
        self.assertTrue(
            issubclass(
                SdBusRequestNameInQueueError,
                SdBusRequestNameError,
            )
        )
        # Test the opposite
        self.assertFalse(
            issubclass(
                SdBusRequestNameAlreadyOwnerError,
                SdBusRequestNameExistsError,
            )
        )
        self.assertFalse(
            issubclass(
                SdBusRequestNameInQueueError,
                SdBusRequestNameExistsError,
            )
        )
        self.assertFalse(
            issubclass(
                SdBusRequestNameInQueueError,
                SdBusRequestNameAlreadyOwnerError,
            )
        )

    async def test_name_exists_async(self) -> None:
        extra_bus = sd_bus_open_user()
        await self.bus.request_name_async(TEST_BUS_NAME, 0)

        with self.assertRaises(SdBusRequestNameExistsError):
            await wait_for(
                extra_bus.request_name_async(TEST_BUS_NAME, 0),
                timeout=1,
            )

    async def test_name_already_async(self) -> None:
        await self.bus.request_name_async(TEST_BUS_NAME, 0)

        with self.assertRaises(SdBusRequestNameAlreadyOwnerError):
            await wait_for(
                self.bus.request_name_async(TEST_BUS_NAME, 0),
                timeout=1,
            )

    async def test_name_queued_async(self) -> None:
        extra_bus = sd_bus_open_user()
        await self.bus.request_name_async(TEST_BUS_NAME, 0)

        with self.assertRaises(SdBusRequestNameInQueueError):
            await wait_for(
                extra_bus.request_name_async(TEST_BUS_NAME, NameQueueFlag),
                timeout=1,
            )

    async def test_name_other_error_async(self) -> None:
        extra_bus = sd_bus_open_user()
        extra_bus.close()

        with self.assertRaises(SdBusLibraryError):
            await wait_for(
                extra_bus.request_name_async(TEST_BUS_NAME, 0),
                timeout=1,
            )

    def test_name_exists_block(self) -> None:
        extra_bus = sd_bus_open_user()
        self.bus.request_name(TEST_BUS_NAME, 0)

        with self.assertRaisesRegex(
            SdBusRequestNameExistsError,
            TEST_BUS_NAME_regex_match,
        ):
            extra_bus.request_name(TEST_BUS_NAME, 0)

    def test_name_already_block(self) -> None:
        self.bus.request_name(TEST_BUS_NAME, 0)

        with self.assertRaisesRegex(
            SdBusRequestNameAlreadyOwnerError,
            TEST_BUS_NAME_regex_match,
        ):
            self.bus.request_name(TEST_BUS_NAME, 0)

    def test_name_queued_block(self) -> None:
        extra_bus = sd_bus_open_user()
        self.bus.request_name(TEST_BUS_NAME, 0)

        with self.assertRaisesRegex(
            SdBusRequestNameInQueueError,
            TEST_BUS_NAME_regex_match,
        ):
            extra_bus.request_name(TEST_BUS_NAME, NameQueueFlag)

    def test_name_other_error_block(self) -> None:
        extra_bus = sd_bus_open_user()
        extra_bus.close()
        with self.assertRaises(SdBusLibraryError):
            extra_bus.request_name(TEST_BUS_NAME, 0)


class TestRequestNameBlock(IsolatedDbusTestCase):
    def test_request_name_replacement(self) -> None:
        extra_bus = sd_bus_open_user()
        extra_bus.request_name(TEST_BUS_NAME, NameAllowReplacementFlag)

        with self.assertRaises(SdBusRequestNameExistsError):
            request_default_bus_name(TEST_BUS_NAME)

        request_default_bus_name(
            TEST_BUS_NAME,
            replace_existing=True,
        )


class TestRequestNameAsync(IsolatedDbusTestCase):
    async def test_request_name_replacement(self) -> None:
        extra_bus = sd_bus_open_user()
        await extra_bus.request_name_async(
            TEST_BUS_NAME,
            NameAllowReplacementFlag,
        )

        with self.assertRaises(SdBusRequestNameExistsError):
            await request_default_bus_name_async(TEST_BUS_NAME)

        await request_default_bus_name_async(
            TEST_BUS_NAME,
            replace_existing=True,
        )

    async def test_request_name_queue(self) -> None:
        extra_bus = sd_bus_open_user()
        await extra_bus.request_name_async(TEST_BUS_NAME, 0)

        with self.assertRaises(SdBusRequestNameInQueueError):
            await request_default_bus_name_async(
                TEST_BUS_NAME,
                queue=True,
            )

        async def catch_owner_changed() -> str:
            dbus = FreedesktopDbus()
            async for name, old, new in dbus.name_owner_changed:
                if name != TEST_BUS_NAME:
                    continue

                if old and new:
                    return new

            raise RuntimeError

        loop = get_running_loop()
        owner_changed_task = loop.create_task(catch_owner_changed())
        await sleep(0)

        extra_bus.close()

        await wait_for(owner_changed_task, timeout=0.5)

        with self.assertRaises(SdBusRequestNameAlreadyOwnerError):
            await request_default_bus_name_async(TEST_BUS_NAME)


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020, 2021 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from importlib.util import find_spec
from unittest import SkipTest, TestCase, main
from unittest.mock import MagicMock, patch

from sdbus.__main__ import generator_main
from sdbus.interface_generator import (
    DbusSigToTyping,
    camel_case_to_snake_case,
    generate_py_file,
    interface_name_to_class,
    interfaces_from_str,
)
from sdbus.unittest import IsolatedDbusTestCase

test_xml = """
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
  "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node name="/com/example/sample_object0">
    <interface name="com.example.SampleInterface0">
      <method name="Frobate">
        <arg name="foo" type="i" direction="in"/>
        <arg name="bar" type="s" direction="out"/>
        <arg name="baz" type="a{us}" direction="out"/>
        <annotation name="org.freedesktop.DBus.Deprecated" value="true"/>
        <annotation name="org.freedesktop.systemd1.Privileged" value="true"/>
      </method>
      <method name="Bazify">
        <arg name="bar" type="(iiu)" direction="in"/>
        <arg name="bar" type="v" direction="out"/>
      </method>
      <method name="Mogrify">
        <arg name="bar" type="(iiav)" direction="in"/>
      </method>
      <signal name="Changed">
        <arg name="new_value" type="b"/>
      </signal>
      <property name="Bar" type="y" access="readwrite"/>
      <property name="FooFoo" type="as" access="read">
        <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal"
        value="false"/>
      </property>
      <property name="BoundBy" type="as" access="read">
        <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal"
        value="const"/>
      </property>
      <property name="FooInvalidates" type="s" access="read">
        <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal"
        value="invalidates"/>
      </property>
    </interface>
    <node name="child_of_sample_object"/>
    <node name="another_child_of_sample_object"/>
</node>
"""


class TestConverter(TestCase):
    def test_camel_to_snake(self) -> None:
        with self.subTest("CamelCase"):
            self.assertEqual(
                'activate_connection',
                camel_case_to_snake_case('ActivateConnection'),
            )

        with self.subTest("Already snake case"):
            self.assertEqual(
                'activate_connection',
                camel_case_to_snake_case('activate_connection'),
            )

        with self.subTest("Upper snake case"):
            self.assertEqual(
                'activate_connection',
                camel_case_to_snake_case('ACTIVATE_CONNECTION'),
            )

    def test_interface_name_to_class(self) -> None:
        self.assertEqual(
            'ComExampleSampleInterface0',
            interface_name_to_class('com.example.SampleInterface0'),
        )

    def test_signature_to_typing(self) -> None:
        with self.subTest('Parse basic'):
            self.assertEqual(
                'str', DbusSigToTyping.typing_basic('s')
            )

            self.assertRaises(
                KeyError, DbusSigToTyping.typing_basic, 'v')

        with self.subTest('Parse variant'):
            self.assertEqual(
                'tuple[str, Any]', DbusSigToTyping.typing_complete('v')
            )

        with self.subTest('Splitter test'):
            self.assertEqual(
                ['v', 'as', '(uisa{sx})', 'h', 'a(ss)', 'a{ss}', 'ay'],
                DbusSigToTyping.split_sig('vas(uisa{sx})ha(ss)a{ss}ay')
            )

        with self.subTest('Parse struct'):
            self.assertEqual(
                DbusSigToTyping.typing_complete('(sx)'),
                'tuple[str, int]',
            )

        with self.subTest('Parse list'):
            self.assertEqual(
                DbusSigToTyping.typing_complete('a(sx)'),
                'list[tuple[str, int]]',
            )

        with self.subTest('Parse dict'):
            self.assertEqual(
                DbusSigToTyping.typing_complete('a{s(xh)}'),
                'dict[str, tuple[int, int]]',
            )

        with self.subTest('Parse signature'):
            self.assertEqual(
                DbusSigToTyping.sig_to_typing('a{s(xh)}'),
                'dict[str, tuple[int, int]]',
            )

            self.assertEqual(
                DbusSigToTyping.sig_to_typing('a{s(xh)}xs'),
                'tuple[dict[str, tuple[int, int]], int, str]',
            )

            self.assertEqual(
                DbusSigToTyping.sig_to_typing('a{s(xh)}xs'),
                'tuple[dict[str, tuple[int, int]], int, str]',
            )

            self.assertEqual(
                DbusSigToTyping.sig_to_typing('as'),
                'list[str]',
            )

            self.assertEqual(
                DbusSigToTyping.sig_to_typing(''),
                'None',
            )

    def test_parsing(self) -> None:
        if find_spec('jinja2') is None:
            raise SkipTest('Jinja2 not installed')

        interfaces_intro = interfaces_from_str(test_xml)

        with self.subTest('Test introspection details'):
            test_interface = interfaces_intro[0]

            for test_property in test_interface.properties:
                if test_property.method_name == 'BoundBy':
                    self.assertEqual(
                        test_property.emits_changed,
                        'const',
                    )
                elif test_property.method_name == 'Bar':
                    self.assertEqual(
                        test_property.emits_changed,
                        True,
                    )
                elif test_property.method_name == 'FooInvalidates':
                    self.assertEqual(
                        test_property.emits_changed,
                        'invalidates',
                    )
                elif test_property.method_name == 'FooFoo':
                    self.assertEqual(
                        test_property.emits_changed,
                        False,
                    )

        generated = generate_py_file(interfaces_intro)
        self.assertIn('flags=DbusPropertyEmitsInvalidationFlag', generated)
        self.assertIn('flags=DbusPropertyConstFlag', generated)


class TestGeneratorAgainstDbus(IsolatedDbusTestCase):
    def setUp(self) -> None:
        if find_spec('jinja2') is None:
            raise SkipTest('Jinja2 not installed')

        super().setUp()

    def test_generate_from_connection(self) -> None:
        with patch("sdbus.__main__.stdout") as stdout_mock:
            generator_main(
                [
                    "gen-from-connection",
                    "org.freedesktop.DBus",
                    "/org/freedesktop/DBus",
                ]
            )

        write_mock: MagicMock = stdout_mock.write
        write_mock.assert_called_once()

        generated_interface = write_mock.call_args.args[0]

        self.assertIn(
            "OrgFreedesktopDBusDebugStatsInterface",
            generated_interface,
        )
        self.assertIn(
            "get_connection_unix_process_id",
            generated_interface,
        )
        self.assertIn(
            "async",
            generated_interface,
        )

    def test_generate_from_connection_blocking(self) -> None:
        with patch("sdbus.__main__.stdout") as stdout_mock:
            generator_main(
                [
                    "gen-from-connection",
                    "--block",
                    "org.freedesktop.DBus",
                    "/org/freedesktop/DBus",
                ]
            )

        write_mock: MagicMock = stdout_mock.write
        write_mock.assert_called_once()

        generated_interface = write_mock.call_args.args[0]

        self.assertNotIn(
            "async",
            generated_interface,
        )
        self.assertIn(
            "dbus_property",
            generated_interface,
        )


INTERFACE_NO_MEMBERS_XML = """
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
  "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node name="/com/example/sample_object0">
    <interface name="com.example.SampleInterface1">
    </interface>
    <node name="child_of_sample_object"/>
    <node name="another_child_of_sample_object"/>
</node>
"""


class TestGeneratorSyntaxCompile(TestCase):
    def setUp(self) -> None:
        if find_spec('jinja2') is None:
            raise SkipTest('Jinja2 not installed')

        super().setUp()

    def test_syntax_compile_async(self) -> None:
        source_code = generate_py_file(
            interfaces_from_str(test_xml),
            do_async=True,
        )
        compile(source_code, filename="<string>", mode="exec")

    def test_syntax_compile_block(self) -> None:
        source_code = generate_py_file(
            interfaces_from_str(test_xml),
            do_async=False,
        )
        compile(source_code, filename="<string>", mode="exec")

    def test_syntax_no_members_interface(self) -> None:

        regular_interface = interfaces_from_str(test_xml)
        no_members_interface = interfaces_from_str(INTERFACE_NO_MEMBERS_XML)

        self.assertFalse(no_members_interface[0].methods)
        self.assertFalse(no_members_interface[0].properties)
        self.assertFalse(no_members_interface[0].signals)

        compile(
            generate_py_file(
                regular_interface + no_members_interface,
                do_async=True,
            ),
            filename="<string>",
            mode="exec",
        )
        compile(
            generate_py_file(
                no_members_interface + regular_interface,
                do_async=True,
            ),
            filename="<string>",
            mode="exec",
        )

        compile(
            generate_py_file(
                regular_interface + no_members_interface,
                do_async=False,
            ),
            filename="<string>",
            mode="exec",
        )
        compile(
            generate_py_file(
                no_members_interface + regular_interface,
                do_async=False,
            ),
            filename="<string>",
            mode="exec",
        )


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020, 2021 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from typing import Optional

from sdbus import (
    DbusInterfaceCommonAsync,
    SdBus,
    dbus_method_async,
    dbus_property_async,
    dbus_signal_async,
)


class FreedesktopDbus(DbusInterfaceCommonAsync,
                      interface_name='org.freedesktop.DBus'):
    """D-Bus daemon."""

    def __init__(self, bus: Optional[SdBus] = None):
        """This is the D-Bus daemon interface. Used for querying D-Bus state.

        D-Bus interface object path and service name is
        predetermined.
        (at ``'org.freedesktop.DBus'``, ``'/org/freedesktop/DBus'``)

        :param SdBus bus:
            Optional D-Bus connection.
            If not passed the default D-Bus will be used.
        """
        super().__init__()
        self._proxify(
            'org.freedesktop.DBus',
            '/org/freedesktop/DBus',
            bus,
        )

    @dbus_method_async('s', method_name='GetConnectionUnixProcessID')
    async def get_connection_pid(self, service_name: str) -> int:
        """Get process ID that owns a specified name.

        :param str service_name: Service name to query.
        :return: PID of name owner
        :raises DbusNameHasNoOwnerError: Nobody owns that name
        """
        raise NotImplementedError

    @dbus_method_async('s', method_name='GetConnectionUnixUser')
    async def get_connection_uid(self, service_name: str) -> int:
        """Get process user ID that owns a specified name.

        :param str service_name: Service name to query.
        :return: User ID of name owner
        :raises DbusNameHasNoOwnerError: Nobody owns that name
        """
        raise NotImplementedError

    @dbus_method_async()
    async def get_id(self) -> str:
        """Returns machine id where bus is run. (stored in ``/etc/machine-id``)

        :return: Machine id
        """
        raise NotImplementedError

    @dbus_method_async('s')
    async def get_name_owner(self, service_name: str) -> str:
        """Returns unique bus name (i.e. ``':1.94'``) for given service name.

        :param str service_name: Service name to query.
        :return: Unique bus name.
        :raises DbusNameHasNoOwnerError: Nobody owns that name
        """
        raise NotImplementedError

    @dbus_method_async()
    async def list_activatable_names(self) -> list[str]:
        """Lists all activatable services names.

        :return: List of all names.
        """
        raise NotImplementedError

    @dbus_method_async()
    async def list_names(self) -> list[str]:
        """List all services and connections currently of the bus.

        :return: List of all current names.
        """
        raise NotImplementedError

    @dbus_method_async('s')
    async def name_has_owner(self, service_name: str) -> bool:
        """
        Return True if someone already owns the name,
        False if nobody does.

        :param str service_name: Service name to query.
        :return: Is the name owned?
        """
        raise NotImplementedError

    @dbus_method_async('su')
    async def start_service_by_name(
            self,
            service_name: str,
            flags: int = 0) -> int:
        """Starts a specified service.

        Flags parameter is not used currently and should be
        omitted or set to 0.

        :param str service_name: Service name to start.
        :param int flags: Not used. Omit or pass 0.
        :return: 1 on success, 2 if already started.
        """
        raise NotImplementedError

    @dbus_property_async('as')
    def features(self) -> list[str]:
        """List of D-Bus daemon features.

        Features include:

        * 'AppArmor' - Messages filtered by AppArmor on this bus.
        * 'HeaderFiltering' - Messages are filtered if they have incorrect \
                              header fields.
        * 'SELinux' - Messages filtered by SELinux on this bus.
        * 'SystemdActivation' - services activated by systemd if their \
                               .service file specifies a D-Bus name.
        """
        raise NotImplementedError

    @dbus_property_async('as')
    def interfaces(self) -> list[str]:
        """Extra D-Bus daemon interfaces"""
        raise NotImplementedError

    @dbus_signal_async('s')
    def name_acquired(self) -> str:
        """Signal when current process acquires a bus name."""
        raise NotImplementedError

    @dbus_signal_async('s')
    def name_lost(self) -> str:
        """Signal when current process loses a bus name."""
        raise NotImplementedError

    @dbus_signal_async('sss')
    def name_owner_changed(self) -> tuple[str, str, str]:
        """Signal when some name on a bus changes owner.

        Is a tuple of:

        * The name that acquired or lost
        * Old owner (by unique bus name) or empty string if no one owned it
        * New owner (by unique bus name) or empty string if no one owns it now
        """
        raise NotImplementedError



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import Event, TimeoutError, wait_for
from contextlib import ExitStack, contextmanager
from operator import setitem
from os import environ, kill
from pathlib import Path
from signal import SIGTERM
from subprocess import DEVNULL
from subprocess import run as subprocess_run
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING
from unittest import IsolatedAsyncioTestCase
from weakref import ref as weak_ref

from .dbus_proxy_async_signal import DbusLocalSignalAsync, DbusProxySignalAsync
from .default_bus import _get_defaul_bus_tls, _set_default_bus_tls
from .sd_bus_internals import SdBusMessage, sd_bus_open_user

if TYPE_CHECKING:
    from collections.abc import Iterator
    from contextlib import AbstractAsyncContextManager
    from typing import Any, Optional, TypeVar, Union

    from .dbus_proxy_async_signal import (
        DbusBoundSignalAsyncBase,
        DbusSignalAsync,
    )
    from .sd_bus_internals import SdBus, SdBusSlot

    T = TypeVar('T')


dbus_config = '''
<busconfig>
  <type>session</type>
  <pidfile>{pidfile_path}</pidfile>
  <auth>EXTERNAL</auth>
  <listen>unix:path={socket_path}</listen>
  <policy context="default">
    <allow send_destination="*" eavesdrop="true"/>
    <allow eavesdrop="true"/>
    <allow own="*"/>
  </policy>
</busconfig>
'''


class DbusSignalRecorderBase:
    def __init__(
        self,
        timeout: Union[int, float],
    ):
        self._timeout = timeout
        self._captured_data: list[Any] = []
        self._ready_event = Event()
        self._callback_method = self._callback

    async def start(self) -> None:
        raise NotImplementedError

    async def stop(self) -> None:
        raise NotImplementedError

    async def __aenter__(self) -> DbusSignalRecorderBase:
        raise NotImplementedError

    async def __aexit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any,
    ) -> None:
        if exc_type is not None:
            return

        try:
            await wait_for(self._ready_event.wait(), timeout=self._timeout)
        except TimeoutError:
            raise AssertionError("D-Bus signal not captured.") from None

    def _callback(self, data: Any) -> None:
        if isinstance(data, SdBusMessage):
            data = data.get_contents()

        self._captured_data.append(data)
        self._ready_event.set()

    @property
    def output(self) -> list[Any]:
        return self._captured_data.copy()


class DbusSignalRecorderRemote(DbusSignalRecorderBase):
    def __init__(
        self,
        timeout: Union[int, float],
        bus: SdBus,
        remote_signal: DbusProxySignalAsync[Any],
    ):
        super().__init__(timeout)
        self._bus = bus
        self._match_slot: Optional[SdBusSlot] = None
        self._remote_signal = remote_signal

    async def __aenter__(self) -> DbusSignalRecorderBase:
        self._match_slot = await self._remote_signal._register_match_slot(
            self._bus,
            self._callback_method,
        )

        return self

    async def __aexit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any,
    ) -> None:
        try:
            await super().__aexit__(exc_type, exc_value, traceback)
        finally:
            if self._match_slot is not None:
                self._match_slot.close()


class DbusSignalRecorderLocal(DbusSignalRecorderBase):
    def __init__(
        self,
        timeout: Union[int, float],
        local_signal: DbusLocalSignalAsync[Any],
    ):
        super().__init__(timeout)
        self._local_signal_ref: weak_ref[DbusSignalAsync[Any]] = (
            weak_ref(local_signal.dbus_signal)
        )

    async def __aenter__(self) -> DbusSignalRecorderBase:
        local_signal = self._local_signal_ref()

        if local_signal is None:
            raise RuntimeError

        local_signal.local_callbacks.add(self._callback_method)
        return self


@contextmanager
def _isolated_dbus(
    dbus_executable_name: str = "dbus-daemon",
) -> Iterator[SdBus]:
    with ExitStack() as exit_stack:
        temp_dir_path = Path(
            exit_stack.enter_context(
                TemporaryDirectory(prefix="python-sdbus-")
            )
        )

        dbus_socket_path = temp_dir_path / "test_dbus.socket"
        pid_path = temp_dir_path / "dbus.pid"
        dbus_config_file = temp_dir_path / "dbus.config"
        dbus_config_file.write_text(
            dbus_config.format(
                socket_path=dbus_socket_path,
                pidfile_path=pid_path
            )
        )

        subprocess_run(
            args=(
                dbus_executable_name,
                '--config-file', dbus_config_file,
                '--fork',
            ),
            stdin=DEVNULL,
            check=True,
        )
        # D-Bus daemon exits once it forks and is initialized.

        dbus_pid = int(pid_path.read_text())
        exit_stack.callback(kill, dbus_pid, SIGTERM)

        old_session_bus_address = environ.get("DBUS_SESSION_BUS_ADDRESS")
        if old_session_bus_address is not None:
            exit_stack.callback(
                setitem,
                environ,
                "DBUS_SESSION_BUS_ADDRESS",
                old_session_bus_address,
            )
        else:
            exit_stack.callback(
                environ.pop,
                "DBUS_SESSION_BUS_ADDRESS",
            )
        environ["DBUS_SESSION_BUS_ADDRESS"] = f"unix:path={dbus_socket_path}"

        old_bus = _get_defaul_bus_tls()
        bus = sd_bus_open_user()
        _set_default_bus_tls(bus)
        exit_stack.callback(_set_default_bus_tls, old_bus)
        yield bus


class IsolatedDbusTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        # TODO: Use enterContext from Python 3.11
        _isolated_dbus_cm = _isolated_dbus()
        self.bus = _isolated_dbus_cm.__enter__()
        self.addCleanup(_isolated_dbus_cm.__exit__, None, None, None)

    def assertDbusSignalEmits(
        self,
        signal: DbusBoundSignalAsyncBase[Any],
        timeout: Union[int, float] = 1,
    ) -> AbstractAsyncContextManager[DbusSignalRecorderBase]:

        if isinstance(signal, DbusLocalSignalAsync):
            return DbusSignalRecorderLocal(timeout, signal)
        elif isinstance(signal, DbusProxySignalAsync):
            return DbusSignalRecorderRemote(timeout, self.bus, signal)
        else:
            raise TypeError("Unknown or unsupported signal class.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2024 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from sdbus import (
    DbusInterfaceCommon,
    DbusInterfaceCommonAsync,
    dbus_method,
    dbus_method_async,
    dbus_property,
    dbus_property_async,
    dbus_signal_async,
)


class TestTypingBlocking(
    DbusInterfaceCommon,
    interface_name="example.com",
):

    @dbus_method(result_signature="as")
    def get_str_list_method(self) -> list[str]:
        raise NotImplementedError

    @dbus_property("as")
    def str_list_property(self) -> list[str]:
        raise NotImplementedError


# These functions are not meant to be executed
# but exist to be type checked.

def check_blocking_interface_method_typing(
    test_interface: TestTypingBlocking,
) -> None:

    should_be_list = test_interface.get_str_list_method()
    should_be_list.append("test")

    for x in should_be_list:
        x.capitalize()


def check_blocking_interface_property_typing(
    test_interface: TestTypingBlocking,
) -> None:

    should_be_list = test_interface.str_list_property
    should_be_list.append("test")

    for x in should_be_list:
        x.capitalize()

    test_interface.str_list_property = ["test", "foobar"]


class TestTypingAsync(
    DbusInterfaceCommonAsync,
    interface_name="example.com",
):

    @dbus_method_async(result_signature="as")
    async def get_str_list_method(self) -> list[str]:
        raise NotImplementedError

    @dbus_property_async("as")
    def str_list_property(self) -> list[str]:
        raise NotImplementedError

    @dbus_signal_async("as")
    def str_list_signal(self) -> list[str]:
        raise NotImplementedError


async def check_async_interface_method_typing(
    test_interface: TestTypingAsync,
) -> None:

    should_be_list = await test_interface.get_str_list_method()
    should_be_list.append("test")

    for x in should_be_list:
        x.capitalize()


async def check_async_interface_property_typing(
    test_interface: TestTypingAsync,
) -> None:

    should_be_list = await test_interface.str_list_property
    should_be_list.append("test")

    for x in should_be_list:
        x.capitalize()

    should_be_list2 = await test_interface.str_list_property.get_async()
    should_be_list2.append("test")

    for x in should_be_list2:
        x.capitalize()

    await test_interface.str_list_property.set_async(["test", "foobar"])


async def check_async_interface_signal_typing(
    test_interface: TestTypingAsync,
) -> None:

    async for ls in test_interface.str_list_signal:
        ls.append("test")
        for x in ls:
            x.capitalize()

    async for ls2 in test_interface.str_list_signal.catch():
        ls2.append("test")
        for x2 in ls2:
            x2.capitalize()

    async for p1, ls3 in test_interface.str_list_signal.catch_anywhere():
        p1.capitalize()
        ls3.append("test")
        for x3 in ls3:
            x3.capitalize()

    async for p2, ls4 in (
        TestTypingAsync.str_list_signal
        .catch_anywhere("org.example")
    ):
        p2.capitalize()
        ls4.append("test")
        for x4 in ls4:
            x4.capitalize()

    test_interface.str_list_signal.emit(["test", "foobar"])


async def check_async_element_class_access_typing() -> None:

    test_list: list[str] = []

    # TODO: Fix dbus async method typing
    # test_list.append(
    #     TestTypingAsync.get_str_list_method.method_name
    # )
    test_list.append(
        TestTypingAsync.str_list_property.property_name
    )
    test_list.append(
        TestTypingAsync.str_list_signal.signal_name
    )



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2025 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from contextvars import copy_context
from unittest import main

from sdbus.unittest import IsolatedDbusTestCase

from sdbus import get_default_bus, sd_bus_open_user, set_context_default_bus


def return_bus_id() -> int:
    return id(get_default_bus())


def set_context_and_return_id() -> int:
    set_context_default_bus(sd_bus_open_user())
    return id(get_default_bus())


class TestDefaultBus(IsolatedDbusTestCase):
    def test_context_bus(self) -> None:
        bus_id = id(get_default_bus())

        self.assertEqual(
            bus_id,
            copy_context().run(return_bus_id),
        )

        self.assertNotEqual(
            bus_id,
            copy_context().run(set_context_and_return_id),
        )

        self.assertEqual(
            bus_id,
            id(get_default_bus()),
        )


if __name__ == "__main__":
    main()
